#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import AnyField
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorRandomPower(TensorDistribution, TensorRandomOperandMixin):
    __slots__ = '_a', '_size'
    _input_fields_ = ['_a']
    _op_type_ = OperandDef.RAND_POWER

    _a = AnyField('a')
    _func_name = 'power'

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super().__init__(_size=size, _state=state, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def a(self):
        return self._a

    def __call__(self, a, chunk_size=None):
        return self.new_tensor([a], None, raw_chunk_size=chunk_size)


def power(random_state, a, size=None, chunk_size=None, gpu=None, dtype=None):
    r"""
    Draws samples in [0, 1] from a power distribution with positive
    exponent a - 1.

    Also known as the power function distribution.

    Parameters
    ----------
    a : float or array_like of floats
        Parameter of the distribution. Should be greater than zero.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``a`` is a scalar.  Otherwise,
        ``mt.array(a).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized power distribution.

    Raises
    ------
    ValueError
        If a < 1.

    Notes
    -----
    The probability density function is

    .. math:: P(x; a) = ax^{a-1}, 0 \le x \le 1, a>0.

    The power function distribution is just the inverse of the Pareto
    distribution. It may also be seen as a special case of the Beta
    distribution.

    It is used, for example, in modeling the over-reporting of insurance
    claims.

    References
    ----------
    .. [1] Christian Kleiber, Samuel Kotz, "Statistical size distributions
           in economics and actuarial sciences", Wiley, 2003.
    .. [2] Heckert, N. A. and Filliben, James J. "NIST Handbook 148:
           Dataplot Reference Manual, Volume 2: Let Subcommands and Library
           Functions", National Institute of Standards and Technology
           Handbook Series, June 2003.
           http://www.itl.nist.gov/div898/software/dataplot/refman2/auxillar/powpdf.pdf

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> a = 5. # shape
    >>> samples = 1000
    >>> s = mt.random.power(a, samples)

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(s.execute(), bins=30)
    >>> x = mt.linspace(0, 1, 100)
    >>> y = a*x**(a-1.)
    >>> normed_y = samples*mt.diff(bins)[0]*y
    >>> plt.plot(x.execute(), normed_y.execute())
    >>> plt.show()

    Compare the power function distribution to the inverse of the Pareto.

    >>> from scipy import stats
    >>> rvs = mt.random.power(5, 1000000)
    >>> rvsp = mt.random.pareto(5, 1000000)
    >>> xx = mt.linspace(0,1,100)
    >>> powpdf = stats.powerlaw.pdf(xx.execute(),5)

    >>> plt.figure()
    >>> plt.hist(rvs.execute(), bins=50, normed=True)
    >>> plt.plot(xx.execute(),powpdf,'r-')
    >>> plt.title('np.random.power(5)')

    >>> plt.figure()
    >>> plt.hist((1./(1.+rvsp)).execute(), bins=50, normed=True)
    >>> plt.plot(xx.execute(),powpdf,'r-')
    >>> plt.title('inverse of 1 + np.random.pareto(5)')

    >>> plt.figure()
    >>> plt.hist((1./(1.+rvsp)).execute(), bins=50, normed=True)
    >>> plt.plot(xx.execute(),powpdf,'r-')
    >>> plt.title('inverse of stats.pareto(5)')
    """
    if dtype is None:
        dtype = np.random.RandomState().power(
            handle_array(a), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorRandomPower(size=size, state=random_state.to_numpy(), gpu=gpu, dtype=dtype)
    return op(a, chunk_size=chunk_size)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import uuid

from .utils import SchedulerActor
from ..utils import log_unhandled

logger = logging.getLogger(__name__)


class SessionActor(SchedulerActor):
    def __init__(self, session_id, **kwargs):
        super().__init__()
        self._session_id = session_id
        self._args = kwargs

        self._tileable_names = dict()
        self._graph_infos = dict()

        self._cluster_info_ref = None
        self._assigner_ref = None
        self._manager_ref = None
        self._graph_refs = dict()
        self._graph_meta_refs = dict()
        self._tileable_to_graph = dict()
        self._mut_tensor_refs = dict()

    @staticmethod
    def gen_uid(session_id):
        return f's:h1:session${session_id}'

    def get_argument(self, key):
        return self._args[key]

    def get_tileable_key(self, name):
        return self._tileable_names[name]

    def get_graph_refs(self):
        return self._graph_refs

    def get_graph_meta_refs(self):
        return self._graph_meta_refs

    def get_graph_ref_by_tileable_key(self, tileable_key):
        return self._tileable_to_graph[tileable_key]

    def post_create(self):
        super().post_create()
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())
        self.set_cluster_info_ref()
        self._manager_ref = self.ctx.actor_ref(SessionManagerActor.default_uid())

        from .assigner import AssignerActor
        assigner_uid = AssignerActor.gen_uid(self._session_id)
        address = self.get_scheduler(assigner_uid)
        self._assigner_ref = self.ctx.create_actor(AssignerActor, uid=assigner_uid, address=address)

    def pre_destroy(self):
        super().pre_destroy()
        self.unset_cluster_info_ref()

        self._manager_ref.delete_session(self._session_id, _tell=True)
        self.ctx.destroy_actor(self._assigner_ref)
        for graph_ref in self._graph_refs.values():
            self.ctx.destroy_actor(graph_ref)
        for mut_tensor_ref in self._mut_tensor_refs.values():
            self.ctx.destroy_actor(mut_tensor_ref)

    def set_graph_info(self, graph_key, **kwargs):
        try:
            info_dict = self._graph_infos[graph_key]
        except KeyError:
            info_dict = self._graph_infos[graph_key] = dict()
        info_dict.update(kwargs)

    def get_graph_infos(self):
        return self._graph_infos

    @log_unhandled
    def submit_tileable_graph(self, serialized_graph, graph_key, target_tileables=None, names=None, compose=True):
        from .graph import GraphActor, GraphMetaActor

        graph_uid = GraphActor.gen_uid(self._session_id, graph_key)

        graph_addr = self.get_scheduler(graph_uid)
        graph_ref = self.ctx.create_actor(GraphActor, self._session_id, graph_key,
                                          serialized_graph, target_tileables=target_tileables,
                                          uid=graph_uid, address=graph_addr)
        self._graph_refs[graph_key] = graph_ref
        self._graph_meta_refs[graph_key] = self.ctx.actor_ref(
            GraphMetaActor.gen_uid(self._session_id, graph_key), address=graph_addr)

        graph_ref.execute_graph(_tell=True, compose=compose)

        for tileable_key in target_tileables or ():
            if tileable_key not in self._tileable_to_graph:
                self._tileable_to_graph[tileable_key] = graph_ref

        if names:
            self._tileable_names.update(zip(names, target_tileables))
        return graph_ref

    @log_unhandled
    def create_mutable_tensor(self, name, shape, dtype, *args, **kwargs):
        from .mutable import MutableTensorActor
        if name in self._mut_tensor_refs:
            raise ValueError(f"The mutable tensor named '{name}' already exists.")
        graph_key = uuid.uuid4()
        mut_tensor_uid = MutableTensorActor.gen_uid(self._session_id, name)
        mut_tensor_addr = self.get_scheduler(mut_tensor_uid)
        mut_tensor_ref = self.ctx.create_actor(MutableTensorActor, self._session_id, name,
                                               shape, dtype, graph_key, uid=mut_tensor_uid,
                                               address=mut_tensor_addr, *args, **kwargs)
        self._mut_tensor_refs[name] = mut_tensor_ref
        return mut_tensor_ref.tensor_meta()

    @log_unhandled
    def get_mutable_tensor(self, name):
        tensor_ref = self._mut_tensor_refs.get(name)
        if tensor_ref is None or tensor_ref.sealed():
            raise ValueError(f"The mutable tensor named '{name}' doesn't exist, or has already been sealed.")
        return tensor_ref.tensor_meta()

    @log_unhandled
    def write_mutable_tensor(self, name, index, value):
        tensor_ref = self._mut_tensor_refs.get(name)
        if tensor_ref is None or tensor_ref.sealed():
            raise ValueError(f"The mutable tensor named '{name}' doesn't exist, or has already been sealed.")
        return tensor_ref.write(index, value)

    @log_unhandled
    def append_chunk_records(self, name, chunk_records):
        tensor_ref = self._mut_tensor_refs.get(name)
        if tensor_ref is None or tensor_ref.sealed():
            raise ValueError(f"The mutable tensor named '{name}' doesn't exist, or has already been sealed.")
        return tensor_ref.append_chunk_records(chunk_records)

    @log_unhandled
    def seal(self, name):
        from .graph import GraphActor, GraphMetaActor
        from .utils import GraphState

        tensor_ref = self._mut_tensor_refs.get(name)
        if tensor_ref is None or tensor_ref.sealed():
            raise ValueError(f"The mutable tensor named '{name}' doesn't exist, or has already been sealed.")

        graph_key_hex, tensor_key, tensor_id, tensor_meta = tensor_ref.seal()
        shape, dtype, chunk_size, chunk_keys, _ = tensor_meta
        graph_key = uuid.UUID(graph_key_hex)

        # Create a GraphActor
        graph_uid = GraphActor.gen_uid(self._session_id, graph_key)
        graph_addr = self.get_scheduler(graph_uid)

        graph_ref = self.ctx.create_actor(GraphActor, self._session_id, graph_key,
                                          serialized_tileable_graph=None,
                                          state=GraphState.SUCCEEDED, final_state=GraphState.SUCCEEDED,
                                          uid=graph_uid, address=graph_addr)
        self._graph_refs[graph_key] = graph_ref
        self._graph_meta_refs[graph_key] = self.ctx.actor_ref(
            GraphMetaActor.gen_uid(self._session_id, graph_key), address=tensor_ref.__getstate__()[0])

        # Add the tensor to the GraphActor
        graph_ref.add_fetch_tileable(tensor_key, tensor_id, shape, dtype, chunk_size, chunk_keys)
        self._tileable_to_graph[tensor_key] = graph_ref

        # Clean up mutable tensor refs.
        actor_ref = self._mut_tensor_refs[name]
        actor_ref.destroy()
        del self._mut_tensor_refs[name]
        return graph_key_hex, tensor_key, tensor_id, tensor_meta

    def graph_state(self, graph_key):
        return self._graph_refs[graph_key].get_state()

    def fetch_result(self, graph_key, tileable_key, check=False):
        # TODO just for test, should move to web handler
        graph_ref = self._graph_refs[graph_key]
        return graph_ref.fetch_tileable_result(tileable_key, _check=check)

    @log_unhandled
    def handle_worker_change(self, adds, removes):
        """
        Receive changes in worker list, collect relevant data losses
        and notify graphs to handle these changes.

        :param adds: endpoints of workers newly added to the cluster
        :param removes: endpoints of workers removed to the cluster
        """
        logger.debug('Worker change detected. adds: %r, removes: %r', adds, removes)

        from .chunkmeta import ChunkMetaActor

        # collect affected chunks
        if removes:
            lost_chunks = set()
            futures = []
            for scheduler in self.get_schedulers():
                ref = self.ctx.actor_ref(ChunkMetaActor.default_uid(), address=scheduler)
                futures.append(ref.remove_workers_in_session(self._session_id, removes, _wait=False))

            for f in futures:
                lost_chunks.update(f.result())
            lost_chunks = list(lost_chunks)
            logger.debug('Meta collection done, %d chunks lost.', len(lost_chunks))
        else:
            lost_chunks = []

        # notify every graph to handle failures
        futures = []
        for ref in self._graph_refs.values():
            futures.append(ref.handle_worker_change(adds, removes, lost_chunks, _wait=False, _tell=True))
        [f.result() for f in futures]


class SessionManagerActor(SchedulerActor):
    def __init__(self):
        super().__init__()
        self._session_refs = dict()

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        self.set_cluster_info_ref()

    def pre_destroy(self):
        self.unset_cluster_info_ref()

    def get_sessions(self):
        return list(self._session_refs.keys())

    def get_session_refs(self):
        return self._session_refs

    @log_unhandled
    def create_session(self, session_id, **kwargs):
        uid = SessionActor.gen_uid(session_id)
        scheduler_address = self.get_scheduler(uid)
        if session_id in self._session_refs:
            return self._session_refs[session_id]
        else:
            session_ref = self.ctx.create_actor(SessionActor, uid=uid, address=scheduler_address,
                                                session_id=session_id, **kwargs)
            self._session_refs[session_id] = session_ref
            return session_ref

    @log_unhandled
    def has_session(self, session_id):
        return session_id in self._session_refs

    @log_unhandled
    def delete_session(self, session_id):
        if session_id in self._session_refs:
            session_ref = self._session_refs[session_id]
            session_ref.destroy()
            del self._session_refs[session_id]

    @log_unhandled
    def broadcast_sessions(self, handler, *args, **kwargs):
        from .assigner import AssignerActor
        futures = []
        for key in self._session_refs.keys():
            ref = self.get_actor_ref(AssignerActor.gen_uid(key))
            futures.append(ref.mark_metrics_expired(_tell=True, _wait=False))
        [f.result() for f in futures]
        for f in futures:
            f.result()

        futures = []
        for ref in self._session_refs.values():
            kwargs.update(dict(_wait=False, _tell=True))
            futures.append(getattr(ref, handler)(*args, **kwargs))
        [f.result() for f in futures]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorFFTMixin, validate_fft, TensorRealFFT


class TensorRFFT(TensorRealFFT, TensorFFTMixin):
    _op_type_ = OperandDef.RFFT

    def __init__(self, n=None, axis=-1, norm=None, dtype=None, **kw):
        super().__init__(_n=n, _axis=axis, _norm=norm, _dtype=dtype, **kw)

    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        if op.n is not None:
            new_shape[op.axis] = op.n
        new_shape[op.axis] = new_shape[op.axis] // 2 + 1
        return tuple(new_shape)


def rfft(a, n=None, axis=-1, norm=None):
    """
    Compute the one-dimensional discrete Fourier Transform for real input.

    This function computes the one-dimensional *n*-point discrete Fourier
    Transform (DFT) of a real-valued array by means of an efficient algorithm
    called the Fast Fourier Transform (FFT).

    Parameters
    ----------
    a : array_like
        Input tensor
    n : int, optional
        Number of points along transformation axis in the input to use.
        If `n` is smaller than the length of the input, the input is cropped.
        If it is larger, the input is padded with zeros. If `n` is not given,
        the length of the input along the axis specified by `axis` is used.
    axis : int, optional
        Axis over which to compute the FFT. If not given, the last axis is
        used.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axis
        indicated by `axis`, or the last one if `axis` is not specified.
        If `n` is even, the length of the transformed axis is ``(n/2)+1``.
        If `n` is odd, the length is ``(n+1)/2``.

    Raises
    ------
    IndexError
        If `axis` is larger than the last axis of `a`.

    See Also
    --------
    mt.fft : For definition of the DFT and conventions used.
    irfft : The inverse of `rfft`.
    fft : The one-dimensional FFT of general (complex) input.
    fftn : The *n*-dimensional FFT.
    rfftn : The *n*-dimensional FFT of real input.

    Notes
    -----
    When the DFT is computed for purely real input, the output is
    Hermitian-symmetric, i.e. the negative frequency terms are just the complex
    conjugates of the corresponding positive-frequency terms, and the
    negative-frequency terms are therefore redundant.  This function does not
    compute the negative frequency terms, and the length of the transformed
    axis of the output is therefore ``n//2 + 1``.

    When ``A = rfft(a)`` and fs is the sampling frequency, ``A[0]`` contains
    the zero-frequency term 0*fs, which is real due to Hermitian symmetry.

    If `n` is even, ``A[-1]`` contains the term representing both positive
    and negative Nyquist frequency (+fs/2 and -fs/2), and must also be purely
    real. If `n` is odd, there is no term at fs/2; ``A[-1]`` contains
    the largest positive frequency (fs/2*(n-1)/n), and is complex in the
    general case.

    If the input `a` contains an imaginary part, it is silently discarded.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.fft.fft([0, 1, 0, 0]).execute()
    array([ 1.+0.j,  0.-1.j, -1.+0.j,  0.+1.j])
    >>> mt.fft.rfft([0, 1, 0, 0]).execute()
    array([ 1.+0.j,  0.-1.j, -1.+0.j])

    Notice how the final element of the `fft` output is the complex conjugate
    of the second element, for real input. For `rfft`, this symmetry is
    exploited to compute only the non-negative frequency terms.

    """
    a = astensor(a)
    validate_fft(a, axis=axis, norm=norm)
    op = TensorRFFT(n=n, axis=axis, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import defaultdict

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import AnyField, StringField
from ..array_utils import as_same_device, device
from ..arithmetic.utils import tree_add
from ..core import TensorOrder
from ..operands import TensorOperand, TensorOperandMixin
from ..utils import decide_unify_split
from .einsumfunc import parse_einsum_input, einsum_path


class TensorEinsum(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.EINSUM

    _subscripts = StringField('subscripts')
    _optimize = AnyField('optimize')
    _order = StringField('order')
    _casting = StringField('casting')

    def __init__(self, subscripts=None, optimize=None, dtype=None, order=None, casting=None, **kw):
        super().__init__(_subscripts=subscripts, _optimize=optimize, _dtype=dtype,
                         _order=order, _casting=casting, **kw)

    @property
    def subscripts(self):
        return self._subscripts

    @property
    def optimize(self):
        return self._optimize

    @property
    def order(self):
        return self._order

    @property
    def casting(self):
        return self._casting

    def __call__(self, input_tensors, shape):
        if self.order in 'KA':
            if any(t.order == TensorOrder.C_ORDER for t in input_tensors):
                order = TensorOrder.C_ORDER
            else:
                order = TensorOrder.F_ORDER
        else:
            if self.order == 'C':
                order = TensorOrder.C_ORDER
            else:
                order = TensorOrder.F_ORDER
        return self.new_tensor(input_tensors, shape=shape, dtype=self.dtype, order=order)

    @classmethod
    def tile(cls, op):
        out_tensor = op.outputs[0]
        input_scripts, output_scripts = op.subscripts.split('->')
        tensor_axes = list(zip(op.inputs, input_scripts.split(',')))

        # rechunk to unify nsplits
        input_nsplits = defaultdict(list)
        for (t, axes) in tensor_axes:
            for splits, ax in zip(t.nsplits, axes):
                input_nsplits[ax].append(splits)
        input_tensors = []
        for (t, axes) in tensor_axes:
            new_nsplits = tuple(decide_unify_split(*input_nsplits[ax]) if t.shape[j] > 1 else (t.shape[j],)
                                for j, ax in enumerate(axes))
            input_tensors.append(t.rechunk(new_nsplits)._inplace_tile())

        tensor_indexes = dict()
        output_axes = defaultdict(list)
        axes_splits = dict()
        tensor_contract_axes = []
        for i, (t, axes) in enumerate(zip(input_tensors, input_scripts.split(','))):
            for j in range(t.ndim):
                if axes[j] in output_scripts:
                    # Record the output tensor's axes and its nsplit.
                    tensor_indexes[axes[j]] = range(len(t.nsplits[j]))
                    output_axes[axes[j]].append((i, j))
            axis_splits = dict((axes[j], t.nsplits[j]) for j in range(t.ndim))
            axes_splits.update(axis_splits)
            tensor_contract_axes.append([ax for ax in axes if ax not in output_scripts])

        out_chunks = []
        output_indexes = [tensor_indexes[ax] for ax in output_scripts]
        for out_idx in itertools.product(*output_indexes):
            all_indexes = [[None] * t.ndim for t in input_tensors]
            tensor_shape = []
            for i, idx in enumerate(out_idx):
                tensor_shape.append(axes_splits[output_scripts[i]][idx])
                for (t_idx, axis) in output_axes[output_scripts[i]]:
                    if input_tensors[t_idx].shape[axis] == 1:
                        all_indexes[t_idx][axis] = 0
                    else:
                        all_indexes[t_idx][axis] = idx
            tensor_shape = tuple(tensor_shape)
            einsum_chunks = []
            contract_axes = [s for s in set(input_scripts.replace(',', '')) if s not in output_scripts]
            for contract_indexes in itertools.product(*[range(len(axes_splits[ax])) for ax in contract_axes]):
                for j, t_contract_axes in enumerate(tensor_contract_axes):
                    for axis in t_contract_axes:
                        axis_index = tensor_axes[j][1].index(axis)
                        all_indexes[j][axis_index] = contract_indexes[contract_axes.index(axis)]
                einsum_op = op.copy().reset_key()
                in_chunks = [t.cix[tuple(indices)] for t, indices in zip(input_tensors, all_indexes)]
                chunk = einsum_op.new_chunk(in_chunks, shape=tensor_shape, order=out_tensor.order)
                einsum_chunks.append(chunk)

            if len(einsum_chunks) == 1:
                c = einsum_chunks[0]
                chunk_op = c.op.copy()
                chunk = chunk_op.new_chunk(c.inputs, shape=c.shape, index=out_idx, order=out_tensor.order)
            else:
                chunk = tree_add(op.dtype, einsum_chunks, out_idx, tensor_shape, sparse=op.sparse)
            out_chunks.append(chunk)

        nsplits = [axes_splits[ax] for ax in output_scripts]
        new_op = op.copy()
        return new_op.new_tensors(input_tensors, out_tensor.shape, chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            if xp is np:
                ctx[op.outputs[0].key] = xp.einsum(op.subscripts, *inputs, optimize=op.optimize, dtype=op.dtype,
                                                   order=op.order, casting=op.casting)
            else:
                # Cupy doesn't support `optimize`, `order` and `casting`.
                ctx[op.outputs[0].key] = xp.einsum(op.subscripts, *inputs, dtype=op.dtype)


def einsum(subscripts, *operands, dtype=None, order='K', casting='safe', optimize=False):
    """
    Evaluates the Einstein summation convention on the operands.

    Using the Einstein summation convention, many common multi-dimensional,
    linear algebraic array operations can be represented in a simple fashion.
    In *implicit* mode `einsum` computes these values.

    In *explicit* mode, `einsum` provides further flexibility to compute
    other array operations that might not be considered classical Einstein
    summation operations, by disabling, or forcing summation over specified
    subscript labels.

    See the notes and examples for clarification.

    :param subscripts: Specifies the subscripts for summation as comma separated list of subscript labels.
    An implicit (classical Einstein summation) calculation is performed unless the explicit indicator ‘->’ is
    included as well as subscript labels of the precise output form.
    :param operands: These are the arrays for the operation.
    :param dtype: If provided, forces the calculation to use the data type specified.
    Note that you may have to also give a more liberal casting parameter to allow the conversions.
    Default is None.
    :param order: Controls the memory layout of the output.
    :param casting: Controls what kind of data casting may occur. Setting this to ‘unsafe’ is not recommended,
     as it can adversely affect accumulations.
    :param optimize: Controls if intermediate optimization should occur.
    :return: The calculation based on the Einstein summation convention.

    The Einstein summation convention can be used to compute
    many multi-dimensional, linear algebraic array operations. `einsum`
    provides a succinct way of representing these.

    A non-exhaustive list of these operations,
    which can be computed by `einsum`, is shown below along with examples:

    * Trace of an array, :py:func:`numpy.trace`.
    * Return a diagonal, :py:func:`numpy.diag`.
    * Array axis summations, :py:func:`numpy.sum`.
    * Transpositions and permutations, :py:func:`numpy.transpose`.
    * Matrix multiplication and dot product, :py:func:`numpy.matmul` :py:func:`numpy.dot`.
    * Vector inner and outer products, :py:func:`numpy.inner` :py:func:`numpy.outer`.
    * Broadcasting, element-wise and scalar multiplication, :py:func:`numpy.multiply`.
    * Tensor contractions, :py:func:`numpy.tensordot`.
    * Chained array operations, in efficient calculation order, :py:func:`numpy.einsum_path`.

    The subscripts string is a comma-separated list of subscript labels,
    where each label refers to a dimension of the corresponding operand.
    Whenever a label is repeated it is summed, so ``mt.einsum('i,i', a, b)``
    is equivalent to :py:func:`mt.inner(a,b) <mars.tensor.inner>`. If a label
    appears only once, it is not summed, so ``mt.einsum('i', a)`` produces a
    view of ``a`` with no changes. A further example ``mt.einsum('ij,jk', a, b)``
    describes traditional matrix multiplication and is equivalent to
    :py:func:`mt.matmul(a,b) <mars.tensor.matmul>`.

    In *implicit mode*, the chosen subscripts are important
    since the axes of the output are reordered alphabetically.  This
    means that ``mt.einsum('ij', a)`` doesn't affect a 2D array, while
    ``mt.einsum('ji', a)`` takes its transpose. Additionally,
    ``mt.einsum('ij,jk', a, b)`` returns a matrix multiplication, while,
    ``mt.einsum('ij,jh', a, b)`` returns the transpose of the
    multiplication since subscript 'h' precedes subscript 'i'.

    In *explicit mode* the output can be directly controlled by
    specifying output subscript labels.  This requires the
    identifier '->' as well as the list of output subscript labels.
    This feature increases the flexibility of the function since
    summing can be disabled or forced when required. The call
    ``mt.einsum('i->', a)`` is like :py:func:`mt.sum(a, axis=-1) <mars.tensor.sum>`,
    and ``mt.einsum('ii->i', a)`` is like :py:func:`mt.diag(a) <mars.tensor.diag>`.
    The difference is that `einsum` does not allow broadcasting by default.
    Additionally ``mt.einsum('ij,jh->ih', a, b)`` directly specifies the
    order of the output subscript labels and therefore returns matrix
    multiplication, unlike the example above in implicit mode.

    To enable and control broadcasting, use an ellipsis.  Default
    NumPy-style broadcasting is done by adding an ellipsis
    to the left of each term, like ``mt.einsum('...ii->...i', a)``.
    To take the trace along the first and last axes,
    you can do ``mt.einsum('i...i', a)``, or to do a matrix-matrix
    product with the left-most indices instead of rightmost, one can do
    ``mt.einsum('ij...,jk...->ik...', a, b)``.

    When there is only one operand, no axes are summed, and no output
    parameter is provided, a view into the operand is returned instead
    of a new array.  Thus, taking the diagonal as ``mt.einsum('ii->i', a)``
    produces a view (changed in version 1.10.0).

    `einsum` also provides an alternative way to provide the subscripts
    and operands as ``einsum(op0, sublist0, op1, sublist1, ..., [sublistout])``.
    If the output shape is not provided in this format `einsum` will be
    calculated in implicit mode, otherwise it will be performed explicitly.
    The examples below have corresponding `einsum` calls with the two
    parameter methods.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.arange(25).reshape(5,5)
    >>> b = mt.arange(5)
    >>> c = mt.arange(6).reshape(2,3)
    Trace of a matrix:
    >>> mt.einsum('ii', a).execute()
    60
    >>> mt.einsum(a, [0,0]).execute()
    60
    Extract the diagonal (requires explicit form):
    >>> mt.einsum('ii->i', a).execute()
    array([ 0,  6, 12, 18, 24])
    >>> mt.einsum(a, [0,0], [0]).execute()
    array([ 0,  6, 12, 18, 24])
    >>> mt.diag(a).execute()
    array([ 0,  6, 12, 18, 24])
    Sum over an axis (requires explicit form):
    >>> mt.einsum('ij->i', a).execute()
    array([ 10,  35,  60,  85, 110])
    >>> mt.einsum(a, [0,1], [0]).execute()
    array([ 10,  35,  60,  85, 110])
    >>> mt.sum(a, axis=1).execute()
    array([ 10,  35,  60,  85, 110])
    For higher dimensional arrays summing a single axis can be done with ellipsis:
    >>> mt.einsum('...j->...', a).execute()
    array([ 10,  35,  60,  85, 110])
    >>> mt.einsum(a, [Ellipsis,1], [Ellipsis]).execute()
    array([ 10,  35,  60,  85, 110])
    Compute a matrix transpose, or reorder any number of axes:
    >>> mt.einsum('ji', c).execute()
    array([[0, 3],
           [1, 4],
           [2, 5]])
    >>> mt.einsum('ij->ji', c).execute()
    array([[0, 3],
           [1, 4],
           [2, 5]])
    >>> mt.einsum(c, [1,0]).execute()
    array([[0, 3],
           [1, 4],
           [2, 5]])
    >>> mt.transpose(c).execute()
    array([[0, 3],
           [1, 4],
           [2, 5]])
    Vector inner products:
    >>> mt.einsum('i,i', b, b).execute()
    30
    >>> mt.einsum(b, [0], b, [0]).execute()
    30
    >>> mt.inner(b,b).execute()
    30
    Matrix vector multiplication:
    >>> mt.einsum('ij,j', a, b).execute()
    array([ 30,  80, 130, 180, 230])
    >>> mt.einsum(a, [0,1], b, [1]).execute()
    array([ 30,  80, 130, 180, 230])
    >>> mt.dot(a, b).execute()
    array([ 30,  80, 130, 180, 230])
    >>> mt.einsum('...j,j', a, b).execute()
    array([ 30,  80, 130, 180, 230])
    Broadcasting and scalar multiplication:
    >>> mt.einsum('..., ...', 3, c).execute()
    array([[ 0,  3,  6],
           [ 9, 12, 15]])
    >>> mt.einsum(',ij', 3, c).execute()
    array([[ 0,  3,  6],
           [ 9, 12, 15]])
    >>> mt.einsum(3, [Ellipsis], c, [Ellipsis]).execute()
    array([[ 0,  3,  6],
           [ 9, 12, 15]])
    >>> mt.multiply(3, c).execute()
    array([[ 0,  3,  6],
           [ 9, 12, 15]])
    Vector outer product:
    >>> mt.einsum('i,j', mt.arange(2)+1, b).execute()
    array([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> mt.einsum(mt.arange(2)+1, [0], b, [1]).execute()
    array([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> mt.outer(mt.arange(2)+1, b).execute()
    array([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    Tensor contraction:
    >>> a = mt.arange(60.).reshape(3,4,5)
    >>> b = mt.arange(24.).reshape(4,3,2)
    >>> mt.einsum('ijk,jil->kl', a, b).execute()
    array([[4400., 4730.],
           [4532., 4874.],
           [4664., 5018.],
           [4796., 5162.],
           [4928., 5306.]])
    >>> mt.einsum(a, [0,1,2], b, [1,0,3], [2,3]).execute()
    array([[4400., 4730.],
           [4532., 4874.],
           [4664., 5018.],
           [4796., 5162.],
           [4928., 5306.]])
    >>> mt.tensordot(a,b, axes=([1,0],[0,1])).execute()
    array([[4400., 4730.],
           [4532., 4874.],
           [4664., 5018.],
           [4796., 5162.],
           [4928., 5306.]])
    Writeable returned arrays (since version 1.10.0):
    >>> a = mt.zeros((3, 3))
    >>> mt.einsum('ii->i', a)[:] = 1
    >>> a.execute()
    array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]])
    Example of ellipsis use:
    >>> a = mt.arange(6).reshape((3,2))
    >>> b = mt.arange(12).reshape((4,3))
    >>> mt.einsum('ki,jk->ij', a, b).execute()
    array([[10, 28, 46, 64],
           [13, 40, 67, 94]])
    >>> mt.einsum('ki,...k->i...', a, b).execute()
    array([[10, 28, 46, 64],
           [13, 40, 67, 94]])
    >>> mt.einsum('k...,jk', a, b).execute()
    array([[10, 28, 46, 64],
           [13, 40, 67, 94]])
    Chained array operations. For more complicated contractions, speed ups
    might be achieved by repeatedly computing a 'greedy' path or pre-computing the
    'optimal' path and repeatedly applying it, using an
    `einsum_path` insertion (since version 1.12.0). Performance improvements can be
    particularly significant with larger arrays:
    >>> a = mt.ones(64).reshape(2,4,8)
    Basic `einsum`: ~1520ms  (benchmarked on 3.1GHz Intel i5.)
    >>> for iteration in range(500):
    ...     _ = mt.einsum('ijk,ilm,njm,nlk,abc->',a,a,a,a,a)
    Sub-optimal `einsum` (due to repeated path calculation time): ~330ms
    >>> for iteration in range(500):
    ...     _ = mt.einsum('ijk,ilm,njm,nlk,abc->',a,a,a,a,a, optimize='optimal')
    Greedy `einsum` (faster optimal path approximation): ~160ms
    >>> for iteration in range(500):
    ...     _ = mt.einsum('ijk,ilm,njm,nlk,abc->',a,a,a,a,a, optimize='greedy')
    Optimal `einsum` (best usage pattern in some use cases): ~110ms
    >>> path = mt.einsum_path('ijk,ilm,njm,nlk,abc->',a,a,a,a,a, optimize='optimal')[0]
    >>> for iteration in range(500):
    ...     _ = mt.einsum('ijk,ilm,njm,nlk,abc->',a,a,a,a,a, optimize=path)

    """
    all_inputs = [subscripts] + list(operands)
    inputs, outputs, operands = parse_einsum_input(all_inputs)
    subscripts = "->".join((inputs, outputs))
    axes_shape = dict()
    for axes, op in zip(inputs.split(','), operands):
        for ax, s in zip(axes, op.shape):
            axes_shape[ax] = s

    if optimize:
        optimize, _ = einsum_path(*all_inputs, optimize=optimize)

    shape = tuple(axes_shape[ax] for ax in outputs)
    op = TensorEinsum(subscripts=subscripts, optimize=optimize, dtype=dtype or operands[0].dtype,
                      order=order, casting=casting)
    return op(operands, shape)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...serialize import KeyField
from ...tiles import TilesError
from ...utils import recursive_tile
from ...tensor import tensor as astensor
from ...tensor.core import TENSOR_TYPE
from ...tensor.utils import decide_unify_split, validate_axis
from ..core import DATAFRAME_TYPE, SERIES_TYPE, IndexValue
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameDot(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DOT

    _lhs = KeyField('lhs')
    _rhs = KeyField('rhs')

    def __init__(self, output_types=None, lhs=None, rhs=None, **kw):
        super().__init__(_output_types=output_types, _lhs=lhs, _rhs=rhs, **kw)

    @property
    def lhs(self):
        return self._lhs

    @property
    def rhs(self):
        return self._rhs

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._lhs = self._inputs[0]
        self._rhs = self._inputs[1]

    def __call__(self, lhs, rhs):
        if not isinstance(rhs, (DATAFRAME_TYPE, SERIES_TYPE)):
            rhs = astensor(rhs)
            test_rhs = rhs
        else:
            test_rhs = rhs.to_tensor()

        test_ret = lhs.to_tensor().dot(test_rhs)
        if test_ret.ndim == 0:
            if isinstance(lhs, SERIES_TYPE) and isinstance(rhs, TENSOR_TYPE):
                # return tensor
                return test_ret
            return self.new_scalar([lhs, rhs], dtype=test_ret.dtype)
        elif test_ret.ndim == 1:
            if lhs.ndim == 1:
                if hasattr(rhs, 'columns_value'):
                    index_value = rhs.columns_value
                else:
                    # tensor
                    length = -1 if np.isnan(rhs.shape[1]) else rhs.shape[1]
                    pd_index = pd.RangeIndex(length)
                    index_value = parse_index(pd_index, store_data=True)
            else:
                assert rhs.ndim == 1
                index_value = lhs.index_value
            return self.new_series([lhs, rhs], shape=test_ret.shape,
                                   dtype=test_ret.dtype, index_value=index_value)
        else:
            if isinstance(rhs, TENSOR_TYPE):
                dtypes = pd.Series(np.repeat(test_ret.dtype, test_ret.shape[1]),
                                   index=pd.RangeIndex(test_ret.shape[1]))
                columns_value = parse_index(dtypes.index, store_data=True)
            else:
                dtypes = pd.Series(np.repeat(test_ret.dtype, test_ret.shape[1]),
                                   index=rhs.columns_value.to_pandas())
                columns_value = rhs.columns_value
            return self.new_dataframe([lhs, rhs], shape=test_ret.shape,
                                      index_value=lhs.index_value,
                                      columns_value=columns_value,
                                      dtypes=dtypes)

    @classmethod
    def _align(cls, lhs, rhs):
        if isinstance(rhs, TENSOR_TYPE):
            # no need to align when rhs is a tensor
            return lhs, rhs

        is_lhs_range_index = False
        if isinstance(lhs, DATAFRAME_TYPE) and \
                isinstance(lhs.columns_value.value, IndexValue.RangeIndex):
            is_lhs_range_index = True
        if isinstance(lhs, SERIES_TYPE) and \
                isinstance(lhs.index_value.value, IndexValue.RangeIndex):
            is_lhs_range_index = True

        is_rhs_range_index = False
        if isinstance(rhs.index_value.value, IndexValue.RangeIndex):
            is_rhs_range_index = True

        if not is_lhs_range_index or not is_rhs_range_index:
            # TODO: e.g. use rhs.loc[lhs.columns_value.to_pandas()]
            # when lhs is a DataFrame and lhs.columns is not a RangeIndex,
            # so does Series
            raise NotImplementedError

        return lhs, rhs

    @classmethod
    def tile(cls, op):
        from ..datasource.from_tensor import dataframe_from_tensor, series_from_tensor

        lhs, rhs = op.lhs, op.rhs
        lhs, rhs = cls._align(lhs, rhs)
        out = op.outputs[0]

        if any(np.isnan(ns) for ns in lhs.nsplits[-1]):
            raise TilesError('lhs has unknown chunk shape on last axis')
        if any(np.isnan(ns) for ns in rhs.nsplits[0]):
            raise TilesError('rhs has unknown chunk shape on first axis')

        nsplit = decide_unify_split(lhs.nsplits[-1], rhs.nsplits[0])
        lhs_axis = validate_axis(lhs.ndim, -1)
        lhs = lhs.rechunk({lhs_axis: nsplit})._inplace_tile()
        rhs = rhs.rechunk({0: nsplit})._inplace_tile()

        # delegate computation to tensor
        lhs_tensor = lhs if isinstance(lhs, TENSOR_TYPE) else lhs.to_tensor()
        rhs_tensor = rhs if isinstance(rhs, TENSOR_TYPE) else rhs.to_tensor()
        ret = lhs_tensor.dot(rhs_tensor)

        if isinstance(out, TENSOR_TYPE):
            pass
        elif ret.ndim == 1:
            ret = series_from_tensor(ret)
            if isinstance(lhs, DATAFRAME_TYPE):
                # lhs DataFrame
                ret._index_value = lhs.index_value
            elif isinstance(rhs, DATAFRAME_TYPE):
                # lhs Series, rhs DataFrame
                ret._index_value = rhs.columns_value
        elif ret.ndim == 2:
            ret = dataframe_from_tensor(ret)
            ret._index_value = lhs.index_value
            if isinstance(rhs, DATAFRAME_TYPE):
                ret._columns_value = rhs.columns_value
                ret._dtypes = rhs.dtypes

        return [recursive_tile(ret)]


def dot(df_or_seris, other):
    op = DataFrameDot(lhs=df_or_seris, rhs=other)
    return op(df_or_seris, other)


dot.__frame_doc__ = """
Compute the matrix multiplication between the DataFrame and other.

This method computes the matrix product between the DataFrame and the
values of an other Series, DataFrame or a numpy array.

It can also be called using ``self @ other`` in Python >= 3.5.

Parameters
----------
other : Series, DataFrame or array-like
    The other object to compute the matrix product with.

Returns
-------
Series or DataFrame
    If other is a Series, return the matrix product between self and
    other as a Serie. If other is a DataFrame or a numpy.array, return
    the matrix product of self and other in a DataFrame of a np.array.

See Also
--------
Series.dot: Similar method for Series.

Notes
-----
The dimensions of DataFrame and other must be compatible in order to
compute the matrix multiplication. In addition, the column names of
DataFrame and the index of other must contain the same values, as they
will be aligned prior to the multiplication.

The dot method for Series computes the inner product, instead of the
matrix product here.

Examples
--------
Here we multiply a DataFrame with a Series.

>>> import mars.tensor as mt
>>> import mars.dataframe as md
>>> df = md.DataFrame([[0, 1, -2, -1], [1, 1, 1, 1]])
>>> s = md.Series([1, 1, 2, 1])
>>> df.dot(s).execute()
0    -4
1     5
dtype: int64

Here we multiply a DataFrame with another DataFrame.

>>> other = md.DataFrame([[0, 1], [1, 2], [-1, -1], [2, 0]])
>>> df.dot(other).execute()
    0   1
0   1   4
1   2   2

Note that the dot method give the same result as @

>>> (df @ other).execute()
    0   1
0   1   4
1   2   2

The dot method works also if other is an np.array.

>>> arr = mt.array([[0, 1], [1, 2], [-1, -1], [2, 0]])
>>> df.dot(arr).execute()
    0   1
0   1   4
1   2   2

Note how shuffling of the objects does not change the result.

>>> s2 = s.reindex([1, 0, 2, 3])
>>> df.dot(s2).execute()
0    -4
1     5
dtype: int64
"""
dot.__series_doc__ = """
Compute the dot product between the Series and the columns of other.

This method computes the dot product between the Series and another
one, or the Series and each columns of a DataFrame, or the Series and
each columns of an array.

It can also be called using `self @ other` in Python >= 3.5.

Parameters
----------
other : Series, DataFrame or array-like
    The other object to compute the dot product with its columns.

Returns
-------
scalar, Series or numpy.ndarray
    Return the dot product of the Series and other if other is a
    Series, the Series of the dot product of Series and each rows of
    other if other is a DataFrame or a numpy.ndarray between the Series
    and each columns of the numpy array.

See Also
--------
DataFrame.dot: Compute the matrix product with the DataFrame.
Series.mul: Multiplication of series and other, element-wise.

Notes
-----
The Series and other has to share the same index if other is a Series
or a DataFrame.

Examples
--------
>>> import mars.tensor as mt
>>> import mars.dataframe as md
>>> s = md.Series([0, 1, 2, 3])
>>> other = md.Series([-1, 2, -3, 4])
>>> s.dot(other).execute()
8
>>> (s @ other).execute()
8
>>> df = md.DataFrame([[0, 1], [-2, 3], [4, -5], [6, 7]])
>>> s.dot(df).execute()
0    24
1    14
dtype: int64
>>> arr = mt.array([[0, 1], [-2, 3], [4, -5], [6, 7]])
>>> s.dot(arr).execute()
array([24, 14])
"""



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import FuseChunkData, FuseChunk
from ..operands import Fuse, FuseChunkMixin, ObjectOperandMixin


class RemoteFuseChunkMixin(ObjectOperandMixin, FuseChunkMixin):
    __slots__ = ()

    def _create_chunk(self, output_idx, index, **kw):
        data = FuseChunkData(_index=index, _op=self, **kw)

        return FuseChunk(data)


class RemoteFuseChunk(RemoteFuseChunkMixin, Fuse):
    pass


class RemoteOperandMixin(ObjectOperandMixin):
    def get_fuse_op_cls(self, _):
        return RemoteFuseChunk



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing

import pandas as pd

from .... import remote as mr
from ....session import Session
from ....utils import ceildiv

try:
    from tsfresh.utilities.distribution import DistributorBaseClass
except ImportError:
    DistributorBaseClass = object


class MarsDistributor(DistributorBaseClass):
    def __init__(self, session=None):
        self._session = session or Session.default_or_local()

    def calculate_best_chunk_size(self, data_length):
        if not hasattr(self._session, 'get_workers_meta'):
            return ceildiv(data_length, multiprocessing.cpu_count())
        else:
            metas = self._session.get_workers_meta()
            n_cores = sum(m['hardware']['cpu_total'] for m in metas.values())
            return ceildiv(data_length, n_cores)

    def distribute(self, func, partitioned_chunks, kwargs):
        def _wrapped_func(*args, **kw):
            # Series.value_counts() may not be able to handle
            if not getattr(pd.Series.value_counts, '_wrapped', False):
                old_value_counts = pd.Series.value_counts

                def _wrapped_value_counts(obj, *args, **kw):
                    try:
                        return old_value_counts(obj, *args, **kw)
                    except ValueError:
                        return old_value_counts(obj.copy(), *args, **kw)

                pd.Series.value_counts = _wrapped_value_counts
                pd.Series.value_counts._wrapped = True

            return func(*args, **kw)

        tasks = []
        for partitioned_chunk in partitioned_chunks:
            tasks.append(mr.spawn(_wrapped_func, args=(partitioned_chunk,), kwargs=kwargs))
        executed = mr.ExecutableTuple(tasks).execute(session=self._session)
        fetched = executed.fetch(session=self._session)
        return [item for results in fetched for item in results]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO

import pandas as pd
import numpy as np

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from ...utils import parse_readable_size, lazy_import, FixedSizeFileObject
from ...serialize import StringField, DictField, ListField, Int32Field, Int64Field, BoolField, AnyField
from ...filesystem import open_file, file_size, glob
from ..arrays import ArrowStringDtype
from ..core import IndexValue
from ..utils import parse_index, build_empty_df, standardize_range_index, to_arrow_dtypes
from ..operands import DataFrameOperand, DataFrameOperandMixin

try:
    from pyarrow import HdfsFile
except ImportError:  # pragma: no cover
    HdfsFile = None


cudf = lazy_import('cudf', globals=globals())


def _find_delimiter(f, block_size=2 ** 16):
    delimiter = b'\n'
    if f.tell() == 0:
        return 0
    while True:
        b = f.read(block_size)
        if not b:
            return f.tell()
        elif delimiter in b:
            return f.tell() - len(b) + b.index(delimiter) + 1


def _find_hdfs_start_end(f, offset, size):
    # As pyarrow doesn't support `readline` operation (https://github.com/apache/arrow/issues/3838),
    # we need to find the start and end of file block manually.

    # Be careful with HdfsFile's seek, it doesn't allow seek beyond EOF.
    loc = min(offset, f.size())
    f.seek(loc)
    start = _find_delimiter(f)
    loc = min(offset + size, f.size())
    f.seek(loc)
    end = _find_delimiter(f)
    return start, end


def _find_chunk_start_end(f, offset, size):
    if HdfsFile is not None and isinstance(f, HdfsFile):
        return _find_hdfs_start_end(f, offset, size)
    f.seek(offset)
    if f.tell() == 0:
        start = 0
    else:
        f.readline()
        start = f.tell()
    f.seek(offset + size)
    f.readline()
    end = f.tell()
    return start, end


class DataFrameReadCSV(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.READ_CSV

    _path = AnyField('path')
    _names = ListField('names')
    _sep = StringField('sep')
    _header = AnyField('header')
    _index_col = Int32Field('index_col')
    _compression = StringField('compression')
    _usecols = ListField('usecols')
    _offset = Int64Field('offset')
    _size = Int64Field('size')
    _nrows = Int64Field('nrows')
    _incremental_index = BoolField('incremental_index')
    _use_arrow_dtype = BoolField('use_arrow_dtype')
    _keep_usecols_order = BoolField('keep_usecols_order')
    _storage_options = DictField('storage_options')

    def __init__(self, path=None, names=None, sep=None, header=None, index_col=None,
                 compression=None, usecols=None, offset=None, size=None, nrows=None,
                 gpu=None, keep_usecols_order=None, incremental_index=None,
                 use_arrow_dtype=None, storage_options=None, **kw):
        super().__init__(_path=path, _names=names, _sep=sep, _header=header,
                         _index_col=index_col, _compression=compression,
                         _usecols=usecols, _offset=offset, _size=size, _nrows=nrows,
                         _gpu=gpu, _incremental_index=incremental_index,
                         _keep_usecols_order=keep_usecols_order,
                         _use_arrow_dtype=use_arrow_dtype,
                         _storage_options=storage_options,
                         _output_types=[OutputType.dataframe], **kw)

    @property
    def path(self):
        return self._path

    @property
    def names(self):
        return self._names

    @property
    def sep(self):
        return self._sep

    @property
    def header(self):
        return self._header

    @property
    def index_col(self):
        return self._index_col

    @property
    def compression(self):
        return self._compression

    @property
    def usecols(self):
        return self._usecols

    @property
    def nrows(self):
        return self._nrows

    @property
    def offset(self):
        return self._offset

    @property
    def size(self):
        return self._size

    @property
    def incremental_index(self):
        return self._incremental_index

    @property
    def use_arrow_dtype(self):
        return self._use_arrow_dtype

    @property
    def keep_usecols_order(self):
        return self._keep_usecols_order

    @property
    def storage_options(self):
        return self._storage_options

    @classmethod
    def _tile_compressed(cls, op):
        # Compression does not support break into small parts
        df = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_op._offset = 0
        chunk_op._size = file_size(op.path)
        shape = df.shape
        new_chunk = chunk_op.new_chunk(None, shape=shape, index=(0, 0), index_value=df.index_value,
                                       columns_value=df.columns_value, dtypes=df.dtypes)
        new_op = op.copy()
        nsplits = ((np.nan,), (df.shape[1],))
        return new_op.new_dataframes(None, df.shape, dtypes=df.dtypes,
                                     index_value=df.index_value,
                                     columns_value=df.columns_value,
                                     chunks=[new_chunk], nsplits=nsplits)

    @classmethod
    def _validate_dtypes(cls, dtypes, is_gpu):
        dtypes = dtypes.to_dict()
        # CuDF doesn't support object type, turn it to 'str'.
        if is_gpu:
            dtypes = dict((n, dt.name if dt != np.dtype('object') else 'str') for n, dt in dtypes.items())
        return dtypes

    @classmethod
    def tile(cls, op):
        if op.compression:
            return cls._tile_compressed(op)

        df = op.outputs[0]
        chunk_bytes = df.extra_params.chunk_bytes
        chunk_bytes = int(parse_readable_size(chunk_bytes)[0])

        dtypes = df.dtypes
        if op.use_arrow_dtype is None and not op.gpu and \
                options.dataframe.use_arrow_dtype:  # pragma: no cover
            # check if use_arrow_dtype set on the server side
            dtypes = to_arrow_dtypes(df.dtypes)

        paths = op.path if isinstance(op.path, (tuple, list)) else glob(op.path, storage_options=op.storage_options)

        out_chunks = []
        index_num = 0
        for path in paths:
            total_bytes = file_size(path)
            offset = 0
            for _ in range(int(np.ceil(total_bytes * 1.0 / chunk_bytes))):
                chunk_op = op.copy().reset_key()
                chunk_op._path = path
                chunk_op._offset = offset
                chunk_op._size = min(chunk_bytes, total_bytes - offset)
                shape = (np.nan, len(dtypes))
                index_value = parse_index(df.index_value.to_pandas(), path, index_num)
                new_chunk = chunk_op.new_chunk(None, shape=shape, index=(index_num, 0), index_value=index_value,
                                               columns_value=df.columns_value, dtypes=dtypes)
                out_chunks.append(new_chunk)
                index_num += 1
                offset += chunk_bytes

        if op.incremental_index and len(out_chunks) > 1 and \
                isinstance(df.index_value._index_value, IndexValue.RangeIndex):
            out_chunks = standardize_range_index(out_chunks)
        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (df.shape[1],))
        return new_op.new_dataframes(None, df.shape, dtypes=dtypes,
                                     index_value=df.index_value,
                                     columns_value=df.columns_value,
                                     chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def _pandas_read_csv(cls, f, op):
        csv_kwargs = op.extra_params.copy()
        out_df = op.outputs[0]
        start, end = _find_chunk_start_end(f, op.offset, op.size)
        f.seek(start)
        b = FixedSizeFileObject(f, end - start)
        if hasattr(out_df, 'dtypes'):
            dtypes = out_df.dtypes
        else:
            # Output will be a Series in some optimize rules.
            dtypes = pd.Series([out_df.dtype], index=[out_df.name])
        if end == start:
            # the last chunk may be empty
            df = build_empty_df(dtypes)
            if op.keep_usecols_order and not isinstance(op.usecols, list):
                # convert to Series, if usecols is a scalar
                df = df[op.usecols]
        else:
            if start == 0:
                # The first chunk contains header
                # As we specify names and dtype, we need to skip header rows
                csv_kwargs['skiprows'] = 1 if op.header == 'infer' else op.header
            if op.usecols:
                usecols = op.usecols if isinstance(op.usecols, list) else [op.usecols]
            else:
                usecols = op.usecols
            if cls._contains_arrow_dtype(dtypes):
                # when keep_default_na is True which is default,
                # will replace null value with np.nan,
                # which will cause failure when converting to arrow string array
                csv_kwargs['keep_default_na'] = False
            df = pd.read_csv(b, sep=op.sep, names=op.names, index_col=op.index_col, usecols=usecols,
                             dtype=dtypes.to_dict(), nrows=op.nrows, **csv_kwargs)
            if op.keep_usecols_order:
                df = df[op.usecols]
        return df

    @classmethod
    def _cudf_read_csv(cls, op):  # pragma: no cover
        if op.usecols:
            usecols = op.usecols if isinstance(op.usecols, list) else [op.usecols]
        else:
            usecols = op.usecols
        csv_kwargs = op.extra_params
        if op.offset == 0:
            df = cudf.read_csv(op.path, byte_range=(op.offset, op.size), sep=op.sep, usecols=usecols, **csv_kwargs)
        else:
            df = cudf.read_csv(op.path, byte_range=(op.offset, op.size), sep=op.sep, names=op.names,
                               usecols=usecols, dtype=cls._validate_dtypes(op.outputs[0].dtypes, op.gpu),
                               nrows=op.nrows, **csv_kwargs)

        if op.keep_usecols_order:
            df = df[op.usecols]
        return df

    @classmethod
    def _contains_arrow_dtype(cls, dtypes):
        return any(isinstance(dtype, ArrowStringDtype) for dtype in dtypes)

    @classmethod
    def execute(cls, ctx, op):
        xdf = cudf if op.gpu else pd
        out_df = op.outputs[0]
        csv_kwargs = op.extra_params.copy()

        with open_file(op.path, compression=op.compression, storage_options=op.storage_options) as f:
            if op.compression is not None:
                # As we specify names and dtype, we need to skip header rows
                csv_kwargs['skiprows'] = 1 if op.header == 'infer' else op.header
                dtypes = cls._validate_dtypes(op.outputs[0].dtypes, op.gpu)
                if cls._contains_arrow_dtype(dtypes.values()):
                    # when keep_default_na is True which is default,
                    # will replace null value with np.nan,
                    # which will cause failure when converting to arrow string array
                    csv_kwargs['keep_default_na'] = False
                df = xdf.read_csv(f, sep=op.sep, names=op.names, index_col=op.index_col,
                                  usecols=op.usecols, dtype=dtypes,
                                  nrows=op.nrows, **csv_kwargs)
                if op.keep_usecols_order:
                    df = df[op.usecols]
            else:
                df = cls._cudf_read_csv(op) if op.gpu else cls._pandas_read_csv(f, op)

        ctx[out_df.key] = df

    def __call__(self, index_value=None, columns_value=None, dtypes=None, chunk_bytes=None):
        shape = (np.nan, len(dtypes))
        return self.new_dataframe(None, shape, dtypes=dtypes, index_value=index_value,
                                  columns_value=columns_value, chunk_bytes=chunk_bytes)


def read_csv(path, names=None, sep=',', index_col=None, compression=None, header='infer',
             dtype=None, usecols=None, nrows=None, chunk_bytes='64M', gpu=None, head_bytes='100k',
             head_lines=None, incremental_index=False, use_arrow_dtype=None,
             storage_options=None, **kwargs):
    r"""
    Read a comma-separated values (csv) file into DataFrame.
    Also supports optionally iterating or breaking of the file
    into chunks.

    Parameters
    ----------
    path : str
        Any valid string path is acceptable. The string could be a URL. Valid
        URL schemes include http, ftp, s3, and file. For file URLs, a host is
        expected. A local file could be: file://localhost/path/to/table.csv.
        If you want to pass in a path object, pandas accepts any ``os.PathLike``.
        By file-like object, we refer to objects with a ``read()`` method, such as
        a file handler (e.g. via builtin ``open`` function) or ``StringIO``.
    sep : str, default ','
        Delimiter to use. If sep is None, the C engine cannot automatically detect
        the separator, but the Python parsing engine can, meaning the latter will
        be used and automatically detect the separator by Python's builtin sniffer
        tool, ``csv.Sniffer``. In addition, separators longer than 1 character and
        different from ``'\s+'`` will be interpreted as regular expressions and
        will also force the use of the Python parsing engine. Note that regex
        delimiters are prone to ignoring quoted data. Regex example: ``'\r\t'``.
    delimiter : str, default ``None``
        Alias for sep.
    header : int, list of int, default 'infer'
        Row number(s) to use as the column names, and the start of the
        data.  Default behavior is to infer the column names: if no names
        are passed the behavior is identical to ``header=0`` and column
        names are inferred from the first line of the file, if column
        names are passed explicitly then the behavior is identical to
        ``header=None``. Explicitly pass ``header=0`` to be able to
        replace existing names. The header can be a list of integers that
        specify row locations for a multi-index on the columns
        e.g. [0,1,3]. Intervening rows that are not specified will be
        skipped (e.g. 2 in this example is skipped). Note that this
        parameter ignores commented lines and empty lines if
        ``skip_blank_lines=True``, so ``header=0`` denotes the first line of
        data rather than the first line of the file.
    names : array-like, optional
        List of column names to use. If the file contains a header row,
        then you should explicitly pass ``header=0`` to override the column names.
        Duplicates in this list are not allowed.
    index_col : int, str, sequence of int / str, or False, default ``None``
      Column(s) to use as the row labels of the ``DataFrame``, either given as
      string name or column index. If a sequence of int / str is given, a
      MultiIndex is used.
      Note: ``index_col=False`` can be used to force pandas to *not* use the first
      column as the index, e.g. when you have a malformed file with delimiters at
      the end of each line.
    usecols : list-like or callable, optional
        Return a subset of the columns. If list-like, all elements must either
        be positional (i.e. integer indices into the document columns) or strings
        that correspond to column names provided either by the user in `names` or
        inferred from the document header row(s). For example, a valid list-like
        `usecols` parameter would be ``[0, 1, 2]`` or ``['foo', 'bar', 'baz']``.
        Element order is ignored, so ``usecols=[0, 1]`` is the same as ``[1, 0]``.
        To instantiate a DataFrame from ``data`` with element order preserved use
        ``pd.read_csv(data, usecols=['foo', 'bar'])[['foo', 'bar']]`` for columns
        in ``['foo', 'bar']`` order or
        ``pd.read_csv(data, usecols=['foo', 'bar'])[['bar', 'foo']]``
        for ``['bar', 'foo']`` order.
        If callable, the callable function will be evaluated against the column
        names, returning names where the callable function evaluates to True. An
        example of a valid callable argument would be ``lambda x: x.upper() in
        ['AAA', 'BBB', 'DDD']``. Using this parameter results in much faster
        parsing time and lower memory usage.
    squeeze : bool, default False
        If the parsed data only contains one column then return a Series.
    prefix : str, optional
        Prefix to add to column numbers when no header, e.g. 'X' for X0, X1, ...
    mangle_dupe_cols : bool, default True
        Duplicate columns will be specified as 'X', 'X.1', ...'X.N', rather than
        'X'...'X'. Passing in False will cause data to be overwritten if there
        are duplicate names in the columns.
    dtype : Type name or dict of column -> type, optional
        Data type for data or columns. E.g. {'a': np.float64, 'b': np.int32,
        'c': 'Int64'}
        Use `str` or `object` together with suitable `na_values` settings
        to preserve and not interpret dtype.
        If converters are specified, they will be applied INSTEAD
        of dtype conversion.
    engine : {'c', 'python'}, optional
        Parser engine to use. The C engine is faster while the python engine is
        currently more feature-complete.
    converters : dict, optional
        Dict of functions for converting values in certain columns. Keys can either
        be integers or column labels.
    true_values : list, optional
        Values to consider as True.
    false_values : list, optional
        Values to consider as False.
    skipinitialspace : bool, default False
        Skip spaces after delimiter.
    skiprows : list-like, int or callable, optional
        Line numbers to skip (0-indexed) or number of lines to skip (int)
        at the start of the file.
        If callable, the callable function will be evaluated against the row
        indices, returning True if the row should be skipped and False otherwise.
        An example of a valid callable argument would be ``lambda x: x in [0, 2]``.
    skipfooter : int, default 0
        Number of lines at bottom of file to skip (Unsupported with engine='c').
    nrows : int, optional
        Number of rows of file to read. Useful for reading pieces of large files.
    na_values : scalar, str, list-like, or dict, optional
        Additional strings to recognize as NA/NaN. If dict passed, specific
        per-column NA values.  By default the following values are interpreted as
        NaN: '', '#N/A', '#N/A N/A', '#NA', '-1.#IND', '-1.#QNAN', '-NaN', '-nan',
        '1.#IND', '1.#QNAN', '<NA>', 'N/A', 'NA', 'NULL', 'NaN', 'n/a',
        'nan', 'null'.
    keep_default_na : bool, default True
        Whether or not to include the default NaN values when parsing the data.
        Depending on whether `na_values` is passed in, the behavior is as follows:
        * If `keep_default_na` is True, and `na_values` are specified, `na_values`
          is appended to the default NaN values used for parsing.
        * If `keep_default_na` is True, and `na_values` are not specified, only
          the default NaN values are used for parsing.
        * If `keep_default_na` is False, and `na_values` are specified, only
          the NaN values specified `na_values` are used for parsing.
        * If `keep_default_na` is False, and `na_values` are not specified, no
          strings will be parsed as NaN.
        Note that if `na_filter` is passed in as False, the `keep_default_na` and
        `na_values` parameters will be ignored.
    na_filter : bool, default True
        Detect missing value markers (empty strings and the value of na_values). In
        data without any NAs, passing na_filter=False can improve the performance
        of reading a large file.
    verbose : bool, default False
        Indicate number of NA values placed in non-numeric columns.
    skip_blank_lines : bool, default True
        If True, skip over blank lines rather than interpreting as NaN values.
    parse_dates : bool or list of int or names or list of lists or dict, default False
        The behavior is as follows:
        * boolean. If True -> try parsing the index.
        * list of int or names. e.g. If [1, 2, 3] -> try parsing columns 1, 2, 3
          each as a separate date column.
        * list of lists. e.g.  If [[1, 3]] -> combine columns 1 and 3 and parse as
          a single date column.
        * dict, e.g. {'foo' : [1, 3]} -> parse columns 1, 3 as date and call
          result 'foo'
        If a column or index cannot be represented as an array of datetimes,
        say because of an unparseable value or a mixture of timezones, the column
        or index will be returned unaltered as an object data type. For
        non-standard datetime parsing, use ``pd.to_datetime`` after
        ``pd.read_csv``. To parse an index or column with a mixture of timezones,
        specify ``date_parser`` to be a partially-applied
        :func:`pandas.to_datetime` with ``utc=True``. See
        :ref:`io.csv.mixed_timezones` for more.
        Note: A fast-path exists for iso8601-formatted dates.
    infer_datetime_format : bool, default False
        If True and `parse_dates` is enabled, pandas will attempt to infer the
        format of the datetime strings in the columns, and if it can be inferred,
        switch to a faster method of parsing them. In some cases this can increase
        the parsing speed by 5-10x.
    keep_date_col : bool, default False
        If True and `parse_dates` specifies combining multiple columns then
        keep the original columns.
    date_parser : function, optional
        Function to use for converting a sequence of string columns to an array of
        datetime instances. The default uses ``dateutil.parser.parser`` to do the
        conversion. Pandas will try to call `date_parser` in three different ways,
        advancing to the next if an exception occurs: 1) Pass one or more arrays
        (as defined by `parse_dates`) as arguments; 2) concatenate (row-wise) the
        string values from the columns defined by `parse_dates` into a single array
        and pass that; and 3) call `date_parser` once for each row using one or
        more strings (corresponding to the columns defined by `parse_dates`) as
        arguments.
    dayfirst : bool, default False
        DD/MM format dates, international and European format.
    cache_dates : bool, default True
        If True, use a cache of unique, converted dates to apply the datetime
        conversion. May produce significant speed-up when parsing duplicate
        date strings, especially ones with timezone offsets.
        .. versionadded:: 0.25.0
    iterator : bool, default False
        Return TextFileReader object for iteration or getting chunks with
        ``get_chunk()``.
    chunksize : int, optional
        Return TextFileReader object for iteration.
        See the `IO Tools docs
        <https://pandas.pydata.org/pandas-docs/stable/io.html#io-chunking>`_
        for more information on ``iterator`` and ``chunksize``.
    compression : {'infer', 'gzip', 'bz2', 'zip', 'xz', None}, default 'infer'
        For on-the-fly decompression of on-disk data. If 'infer' and
        `filepath_or_buffer` is path-like, then detect compression from the
        following extensions: '.gz', '.bz2', '.zip', or '.xz' (otherwise no
        decompression). If using 'zip', the ZIP file must contain only one data
        file to be read in. Set to None for no decompression.
    thousands : str, optional
        Thousands separator.
    decimal : str, default '.'
        Character to recognize as decimal point (e.g. use ',' for European data).
    lineterminator : str (length 1), optional
        Character to break file into lines. Only valid with C parser.
    quotechar : str (length 1), optional
        The character used to denote the start and end of a quoted item. Quoted
        items can include the delimiter and it will be ignored.
    quoting : int or csv.QUOTE_* instance, default 0
        Control field quoting behavior per ``csv.QUOTE_*`` constants. Use one of
        QUOTE_MINIMAL (0), QUOTE_ALL (1), QUOTE_NONNUMERIC (2) or QUOTE_NONE (3).
    doublequote : bool, default ``True``
       When quotechar is specified and quoting is not ``QUOTE_NONE``, indicate
       whether or not to interpret two consecutive quotechar elements INSIDE a
       field as a single ``quotechar`` element.
    escapechar : str (length 1), optional
        One-character string used to escape other characters.
    comment : str, optional
        Indicates remainder of line should not be parsed. If found at the beginning
        of a line, the line will be ignored altogether. This parameter must be a
        single character. Like empty lines (as long as ``skip_blank_lines=True``),
        fully commented lines are ignored by the parameter `header` but not by
        `skiprows`. For example, if ``comment='#'``, parsing
        ``#empty\na,b,c\n1,2,3`` with ``header=0`` will result in 'a,b,c' being
        treated as the header.
    encoding : str, optional
        Encoding to use for UTF when reading/writing (ex. 'utf-8'). `List of Python
        standard encodings
        <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ .
    dialect : str or csv.Dialect, optional
        If provided, this parameter will override values (default or not) for the
        following parameters: `delimiter`, `doublequote`, `escapechar`,
        `skipinitialspace`, `quotechar`, and `quoting`. If it is necessary to
        override values, a ParserWarning will be issued. See csv.Dialect
        documentation for more details.
    error_bad_lines : bool, default True
        Lines with too many fields (e.g. a csv line with too many commas) will by
        default cause an exception to be raised, and no DataFrame will be returned.
        If False, then these "bad lines" will dropped from the DataFrame that is
        returned.
    warn_bad_lines : bool, default True
        If error_bad_lines is False, and warn_bad_lines is True, a warning for each
        "bad line" will be output.
    delim_whitespace : bool, default False
        Specifies whether or not whitespace (e.g. ``' '`` or ``'    '``) will be
        used as the sep. Equivalent to setting ``sep='\s+'``. If this option
        is set to True, nothing should be passed in for the ``delimiter``
        parameter.
    low_memory : bool, default True
        Internally process the file in chunks, resulting in lower memory use
        while parsing, but possibly mixed type inference.  To ensure no mixed
        types either set False, or specify the type with the `dtype` parameter.
        Note that the entire file is read into a single DataFrame regardless,
        use the `chunksize` or `iterator` parameter to return the data in chunks.
        (Only valid with C parser).
    float_precision : str, optional
        Specifies which converter the C engine should use for floating-point
        values. The options are `None` for the ordinary converter,
        `high` for the high-precision converter, and `round_trip` for the
        round-trip converter.
    chunk_bytes: int, float or str, optional
        Number of chunk bytes.
    gpu: bool, default False
        If read into cudf DataFrame.
    head_bytes: int, float or str, optional
        Number of bytes to use in the head of file, mainly for data inference.
    head_lines: int, optional
        Number of lines to use in the head of file, mainly for data inference.
    incremental_index: bool, default False
        Create a new RangeIndex if csv doesn't contain index columns.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns.
    storage_options: dict, optional
        Options for storage connection.

    Returns
    -------
    DataFrame
        A comma-separated values (csv) file is returned as two-dimensional
        data structure with labeled axes.

    See Also
    --------
    to_csv : Write DataFrame to a comma-separated values (csv) file.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> md.read_csv('data.csv')  # doctest: +SKIP
    """
    # infer dtypes and columns
    if isinstance(path, (list, tuple)):
        file_path = path[0]
    else:
        file_path = glob(path)[0]
    with open_file(file_path, compression=compression, storage_options=storage_options) as f:
        if head_lines is not None:
            b = b''.join([f.readline() for _ in range(head_lines)])
        else:
            head_bytes = int(parse_readable_size(head_bytes)[0])
            head_start, head_end = _find_chunk_start_end(f, 0, head_bytes)
            f.seek(head_start)
            b = f.read(head_end - head_start)
        mini_df = pd.read_csv(BytesIO(b), sep=sep, index_col=index_col, dtype=dtype,
                              names=names, header=header)

    if isinstance(mini_df.index, pd.RangeIndex):
        index_value = parse_index(pd.RangeIndex(-1))
    else:
        index_value = parse_index(mini_df.index)
    columns_value = parse_index(mini_df.columns, store_data=True)
    if index_col and not isinstance(index_col, int):
        index_col = list(mini_df.columns).index(index_col)
    names = list(mini_df.columns)
    op = DataFrameReadCSV(path=path, names=names, sep=sep, header=header, index_col=index_col,
                          usecols=usecols, compression=compression, gpu=gpu,
                          incremental_index=incremental_index, use_arrow_dtype=use_arrow_dtype,
                          storage_options=storage_options,
                          **kwargs)
    chunk_bytes = chunk_bytes or options.chunk_store_limit
    dtypes = mini_df.dtypes
    if use_arrow_dtype is None:
        use_arrow_dtype = options.dataframe.use_arrow_dtype
    if not gpu and use_arrow_dtype:
        dtypes = to_arrow_dtypes(dtypes, test_df=mini_df)
    ret = op(index_value=index_value, columns_value=columns_value,
             dtypes=dtypes, chunk_bytes=chunk_bytes)
    if nrows is not None:
        return ret.head(nrows)
    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections.abc import Iterable

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import ValueType, KeyField, AnyField, TupleField, BoolField
from ...utils import recursive_tile
from ..utils import validate_axis
from ..array_utils import device, as_same_device
from ..operands import TensorHasInput, TensorOperandMixin
from ..arithmetic import sqrt
from ..datasource import tensor as astensor
from .svd import svd


class TensorNorm(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.NORM

    _input = KeyField('input')
    _ord = AnyField('ord')
    _axis = TupleField('axis', ValueType.int32)
    _keepdims = BoolField('keepdims')

    def __init__(self, ord=None, axis=None, keepdims=None, dtype=None, sparse=False, **kw):
        super().__init__(_ord=ord, _axis=axis, _keepdims=keepdims, _dtype=dtype,
                         _sparse=sparse, **kw)

    @property
    def ord(self):
        return getattr(self, '_ord', None)

    @property
    def axis(self):
        return self._axis

    @property
    def keepdims(self):
        return self._keepdims

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, x):
        r = x.astype(self.dtype)
        shape = self._norm(r, self._ord, self._axis, self._keepdims).shape
        return self.new_tensor([x], shape)

    @classmethod
    def tile(cls, op):
        x = astensor(op.input)
        axis = op.axis
        ord = op.ord
        keepdims = op.keepdims

        axis_chunk_shapes = tuple(x.chunk_shape[i] for i in axis)
        can_apply_norm = all(s == 1 for s in axis_chunk_shapes)

        if can_apply_norm:
            axis_set = set(axis)
            get_shape = lambda shape: tuple(s if i not in axis_set else 1 for i, s in enumerate(shape)
                                            if i not in axis_set or keepdims)

            out_chunk_shape = get_shape(x.chunk_shape)
            out_chunks = []
            for idx in itertools.product(*[range(s) for s in out_chunk_shape]):
                idx_iter = iter(idx)
                in_idx = tuple(0 if i in axis_set and not keepdims else next(idx_iter)
                               for i in range(x.ndim))

                c = x.cix[in_idx]
                chunk_op = op.copy().reset_key()
                out_chunk = chunk_op.new_chunk([c], shape=get_shape(c.shape), index=idx)
                out_chunks.append(out_chunk)

            nsplits = [tuple(c.shape[i] for c in out_chunks
                             if all(idx == 0 for j, idx in enumerate(c.index) if j != i))
                       for i in range(len(out_chunks[0].shape))]
            new_op = op.copy()
            return new_op.new_tensors(op.inputs, op.outputs[0].shape, chunks=out_chunks, nsplits=nsplits)

        r = cls._norm(x.astype(op.outputs[0].dtype), ord, axis, keepdims)
        recursive_tile(r)
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape, chunks=r.chunks, nsplits=r.nsplits)

    @staticmethod
    def _norm(r, ord, axis, keepdims):
        if ord is None:
            return sqrt((abs(r) ** 2).sum(axis=axis, keepdims=keepdims))
        elif ord == 'nuc':
            if len(axis) == 1:
                raise ValueError('Invalid norm order for vectors.')
            return svd(r)[1][np.newaxis].sum(keepdims=keepdims)
        elif ord == np.inf:
            if r.ndim > 2:
                raise ValueError('Improper number of dimensions to norm.')
            r = abs(r)
            if len(axis) == 1:
                return r.max(axis=axis, keepdims=keepdims)
            else:
                return r.sum(axis=axis[1], keepdims=keepdims).max(keepdims=keepdims)
        elif ord == -np.inf:
            if r.ndim > 2:
                raise ValueError('Improper number of dimensions to norm.')
            r = abs(r)
            if len(axis) == 1:
                return r.min(axis=axis, keepdims=keepdims)
            else:
                return r.sum(axis=axis[1], keepdims=keepdims).min(keepdims=keepdims)
        elif ord == 0:
            if r.ndim > 2:
                raise ValueError('Improper number of dimensions to norm.')
            if len(axis) == 2:
                raise ValueError('Invalid norm order for matrices.')
            return (r != 0).astype(r.dtype).sum(axis=axis, keepdims=keepdims)
        elif ord == 1:
            if r.ndim > 2:
                raise ValueError('Improper number of dimensions to norm.')
            r = abs(r)
            if len(axis) == 1:
                return r.sum(axis=axis, keepdims=keepdims)
            else:
                return r.sum(axis=axis[0], keepdims=keepdims).max(keepdims=keepdims)
        elif ord == -1 and len(axis) == 2:
            if r.ndim > 2:
                raise ValueError('Improper number of dimensions to norm.')
            return abs(r).sum(axis=axis[0], keepdims=keepdims).min(keepdims=keepdims)
        elif ord == 2 and len(axis) == 2:
            return svd(r)[1][np.newaxis].max(keepdims=keepdims)
        elif ord == -2 and len(axis) == 2:
            return svd(r)[1][np.newaxis].min(keepdims=keepdims)
        else:
            if len(axis) == 2:
                raise ValueError('Invalid norm order for matrices.')

            return (abs(r) ** ord).sum(axis=axis, keepdims=keepdims) ** (1.0 / ord)

    @classmethod
    def execute(cls, ctx, op):
        (x,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            ctx[op.outputs[0].key] = xp.linalg.norm(x, ord=op.ord, axis=op.axis,
                                                    keepdims=op.keepdims)


def norm(x, ord=None, axis=None, keepdims=False):
    r"""
    Matrix or vector norm.

    This function is able to return one of eight different matrix norms,
    or one of an infinite number of vector norms (described below), depending
    on the value of the ``ord`` parameter.

    Parameters
    ----------
    x : array_like
        Input tensor.  If `axis` is None, `x` must be 1-D or 2-D.
    ord : {non-zero int, inf, -inf, 'fro', 'nuc'}, optional
        Order of the norm (see table under ``Notes``). inf means mars tensor's
        `inf` object.
    axis : {int, 2-tuple of ints, None}, optional
        If `axis` is an integer, it specifies the axis of `x` along which to
        compute the vector norms.  If `axis` is a 2-tuple, it specifies the
        axes that hold 2-D matrices, and the matrix norms of these matrices
        are computed.  If `axis` is None then either a vector norm (when `x`
        is 1-D) or a matrix norm (when `x` is 2-D) is returned.
    keepdims : bool, optional
        If this is set to True, the axes which are normed over are left in the
        result as dimensions with size one.  With this option the result will
        broadcast correctly against the original `x`.

    Returns
    -------
    n : float or Tensor
        Norm of the matrix or vector(s).

    Notes
    -----
    For values of ``ord <= 0``, the result is, strictly speaking, not a
    mathematical 'norm', but it may still be useful for various numerical
    purposes.

    The following norms can be calculated:

    =====  ============================  ==========================
    ord    norm for matrices             norm for vectors
    =====  ============================  ==========================
    None   Frobenius norm                2-norm
    'fro'  Frobenius norm                --
    'nuc'  nuclear norm                  --
    inf    max(sum(abs(x), axis=1))      max(abs(x))
    -inf   min(sum(abs(x), axis=1))      min(abs(x))
    0      --                            sum(x != 0)
    1      max(sum(abs(x), axis=0))      as below
    -1     min(sum(abs(x), axis=0))      as below
    2      2-norm (largest sing. value)  as below
    -2     smallest singular value       as below
    other  --                            sum(abs(x)**ord)**(1./ord)
    =====  ============================  ==========================

    The Frobenius norm is given by [1]_:

        :math:`||A||_F = [\\sum_{i,j} abs(a_{i,j})^2]^{1/2}`

    The nuclear norm is the sum of the singular values.

    References
    ----------
    .. [1] G. H. Golub and C. F. Van Loan, *Matrix Computations*,
           Baltimore, MD, Johns Hopkins University Press, 1985, pg. 15

    Examples
    --------
    >>> from mars.tensor import linalg as LA
    >>> import mars.tensor as mt
    >>> a = mt.arange(9) - 4
    >>> a.execute()
    array([-4, -3, -2, -1,  0,  1,  2,  3,  4])
    >>> b = a.reshape((3, 3))
    >>> b.execute()
    array([[-4, -3, -2],
           [-1,  0,  1],
           [ 2,  3,  4]])

    >>> LA.norm(a).execute()
    7.745966692414834
    >>> LA.norm(b).execute()
    7.745966692414834
    >>> LA.norm(b, 'fro').execute()
    7.745966692414834
    >>> LA.norm(a, mt.inf).execute()
    4.0
    >>> LA.norm(b, mt.inf).execute()
    9.0
    >>> LA.norm(a, -mt.inf).execute()
    0.0
    >>> LA.norm(b, -mt.inf).execute()
    2.0

    >>> LA.norm(a, 1).execute()
    20.0
    >>> LA.norm(b, 1).execute()
    7.0
    >>> LA.norm(a, -1).execute()
    0.0
    >>> LA.norm(b, -1).execute()
    6.0
    >>> LA.norm(a, 2).execute()
    7.745966692414834
    >>> LA.norm(b, 2).execute()
    7.3484692283495345

    >>> LA.norm(a, -2).execute()
    0.0
    >>> LA.norm(b, -2).execute()
    4.351066026358965e-18
    >>> LA.norm(a, 3).execute()
    5.8480354764257312
    >>> LA.norm(a, -3).execute()
    0.0

    Using the `axis` argument to compute vector norms:

    >>> c = mt.array([[ 1, 2, 3],
    ...               [-1, 1, 4]])
    >>> LA.norm(c, axis=0).execute()
    array([ 1.41421356,  2.23606798,  5.        ])
    >>> LA.norm(c, axis=1).execute()
    array([ 3.74165739,  4.24264069])
    >>> LA.norm(c, ord=1, axis=1).execute()
    array([ 6.,  6.])

    Using the `axis` argument to compute matrix norms:

    >>> m = mt.arange(8).reshape(2,2,2)
    >>> LA.norm(m, axis=(1,2)).execute()
    array([  3.74165739,  11.22497216])
    >>> LA.norm(m[0, :, :]).execute(), LA.norm(m[1, :, :]).execute()
    (3.7416573867739413, 11.224972160321824)

    """
    x = astensor(x)
    ndim = x.ndim

    if ord == 'fro':
        ord = None
    if axis is not None:
        if isinstance(axis, Iterable):
            axis = tuple(validate_axis(ndim, a) for a in axis)
        else:
            axis = (validate_axis(ndim, axis),)
    else:
        axis = tuple(range(x.ndim))

    op = TensorNorm(ord=ord, axis=axis, keepdims=keepdims,
                    dtype=np.result_type(x.dtype, np.float_), sparse=x.issparse())
    return op(x)



# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import textwrap
from urllib.parse import urlparse

from ...utils import parse_readable_size


def _remove_nones(cfg):
    return dict((k, v) for k, v in cfg.items() if v is not None)


def _get_local_app_module(mod_name):
    return __name__.rsplit('.', 1)[0] + '.' + mod_name.rsplit('.', 1)[-1]


class SecurityConfig:
    def __init__(self, cert_file=None, key_file=None):
        self._cert_file = cert_file
        self._key_file = key_file

    def build(self):
        return dict(cert_file=self._cert_file, key_file=self._key_file)


class AppFileConfig:
    def __init__(self, source, file_type=None, visibility=None,
                 size=None, timestamp=None):
        self._source = source
        self._file_type = file_type
        self._visibility = visibility
        self._size = size
        self._timestamp = timestamp

    def build(self):
        if all(v is None for v in (self._file_type, self._visibility, self._size, self._timestamp)):
            return self._source
        else:
            return _remove_nones(dict(
                source=self._source, type=self._file_type, visibility=self._visibility,
                size=self._size, timestamp=self._timestamp
            ))


class AppContainerConfig:
    def __init__(self, cpu=None, memory=None, env=None, files=None, script=None):
        self._cpu = cpu

        if memory is not None:
            real_mem, is_percent = parse_readable_size(memory)
            assert not is_percent
            self._memory = real_mem
        else:
            self._memory = None

        self._env = env
        self._script = script
        self._files = files

        self.add_default_envs()

    def build_script(self):
        return self._script

    def add_default_envs(self):
        pass

    def add_env(self, k, v):
        if self._env is None:
            self._env = dict()
        self._env[k] = v

    def build(self):
        return _remove_nones(dict(
            resources=dict(
                vcores=self._cpu,
                memory=f'{self._memory // 1024 ** 2} MiB' if self._memory else None,
            ),
            env=self._env,
            script=self.build_script(),
            files=dict((k, v.build()) for k, v in self._files.items()) if self._files else None,
        ))


class AppMasterConfig(AppContainerConfig):
    def __init__(self, security=None, **kwargs):
        super().__init__(**kwargs)
        self._security = security

    def build(self):
        d = super().build()
        if self._security is not None:
            d['security'] = self._security.build()
        return d


class AppServiceConfig(AppContainerConfig):
    def __init__(self, instances=1, depends=None, allow_failures=False,
                 max_restarts=0, **kwargs):
        super().__init__(**kwargs)
        if isinstance(depends, str):
            depends = [depends]

        self._allow_failures = allow_failures
        self._depends = depends or []
        self._max_restarts = max_restarts
        self._instances = instances

    def build(self):
        d = super().build()
        d.update(dict(
            instances=self._instances,
            depends=self._depends,
            allow_failures=self._allow_failures,
            max_restarts=self._max_restarts,
        ))
        return d


class MarsServiceConfig(AppServiceConfig):
    service_name = None

    def __init__(self, environment, modules=None, cmd_tmpl=None, cpu=None, memory=None,
                 log_config=None, extra_args=None, **kwargs):
        files = kwargs.pop('files', dict())
        kwargs['files'] = files

        parsed = urlparse(environment)
        self._env_scheme = parsed.scheme

        if parsed.scheme:
            import mars
            self._source_path = os.path.dirname(os.path.dirname(os.path.abspath(mars.__file__)))

            self._env_path = environment[len(parsed.scheme) + 3:]
            self._path_environ = os.environ['PATH']
        else:
            self._source_path = None
            self._env_path = environment
            self._path_environ = None

        self._cmd_tmpl = cmd_tmpl or '"{executable}"'
        if not self._env_scheme:
            files['mars_env'] = AppFileConfig(environment)

        self._log_config = log_config
        if log_config:
            files['logging.conf'] = AppFileConfig(log_config)

        self._modules = modules.split(',') if isinstance(modules, str) else modules

        self._extra_args = extra_args or ''

        cpu = cpu or 1
        memory = memory or '1 GiB'
        super().__init__(cpu=cpu, memory=memory, **kwargs)

    def add_default_envs(self):
        if self._cpu:
            self.add_env('MKL_NUM_THREADS', str(self._cpu))
            self.add_env('MARS_CPU_TOTAL', str(self._cpu))
            self.add_env('MARS_USE_PROCESS_STAT', '1')

        if self._memory:
            self.add_env('MARS_MEMORY_TOTAL', str(int(self._memory)))

        if self._modules:
            self.add_env('MARS_LOAD_MODULES', ','.join(self._modules))

        if self._path_environ:
            self.add_env('MARS_YARN_PATH', self._path_environ)

        if self._source_path:
            self.add_env('MARS_SOURCE_PATH', self._source_path)

    def build_script(self):
        bash_lines = [textwrap.dedent("""
        #!/bin/bash
        if [[ "$YARN_CONTAINER_RUNTIME_TYPE" == "docker" ]]; then
          export MARS_USE_CGROUP_STAT=1
        else
          export MARS_USE_PROCESS_STAT=1
        fi
        if [[ -n $MARS_SOURCE_PATH ]]; then export PYTHONPATH=$PYTHONPATH:$MARS_SOURCE_PATH; fi
        if [[ -n $MARS_YARN_PATH ]]; then export PATH=$MARS_YARN_PATH:$PATH; fi
        """).strip()]

        if not self._env_scheme:
            bash_lines.append('source mars_env/bin/activate')
            python_executable = 'mars_env/bin/python'
        elif self._env_scheme == 'conda':
            bash_lines.append(f'conda activate "{self._env_path}"')
            python_executable = 'python'
        elif self._env_scheme == 'venv':
            bash_lines.append(f'source "{self._env_path}/bin/activate"')
            python_executable = self._env_path + '/bin/python'
        else:  # pragma: no cover
            python_executable = self._env_path

        cmd = self._cmd_tmpl.format(executable=python_executable)
        bash_lines.append(f'{cmd} -m {_get_local_app_module(self.service_name)} {self._extra_args}')
        return '\n'.join(bash_lines) + '\n'


class MarsSchedulerConfig(MarsServiceConfig):
    service_name = 'mars.scheduler'


class MarsWorkerConfig(MarsServiceConfig):
    service_name = 'mars.worker'

    def __init__(self, environment, worker_cache_mem=None, spill_dirs=None, **kwargs):
        kwargs['depends'] = MarsSchedulerConfig.service_name
        super().__init__(environment, **kwargs)

        if worker_cache_mem:
            self.add_env('MARS_CACHE_MEM_SIZE', worker_cache_mem)

        if spill_dirs:
            self.add_env('MARS_SPILL_DIRS',
                         spill_dirs if isinstance(spill_dirs, str) else ':'.join(spill_dirs))


class MarsWebConfig(MarsServiceConfig):
    service_name = 'mars.web'

    def __init__(self, environment, **kwargs):
        kwargs['depends'] = MarsSchedulerConfig.service_name
        super().__init__(environment, **kwargs)


class MarsApplicationConfig:
    def __init__(self, name=None, queue=None, file_systems=None, master=None,
                 scheduler_config=None, worker_config=None, web_config=None):
        self._name = name
        self._queue = queue or 'default'
        self._file_systems = file_systems or []
        self._master = master or AppMasterConfig(cpu=1, memory='512 MiB')
        self._scheduler_config = scheduler_config
        self._worker_config = worker_config
        self._web_config = web_config

    def build(self):
        services = _remove_nones({
            MarsSchedulerConfig.service_name: self._scheduler_config.build() if self._scheduler_config else None,
            MarsWorkerConfig.service_name: self._worker_config.build() if self._worker_config else None,
            MarsWebConfig.service_name: self._web_config.build() if self._web_config else None,
        })
        return dict(
            name=self._name,
            queue=self._queue,
            file_systems=self._file_systems,
            master=self._master.build() if self._master else None,
            services=services,
        )



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from ... import promise
from ...config import options
from ...errors import StorageFull, StorageDataExists
from ...serialize import dataserializer
from ..dataio import ArrowBufferIO
from .core import StorageHandler, BytesStorageMixin, ObjectStorageMixin, \
    SpillableStorageMixin, BytesStorageIO, DataStorageDevice, wrap_promised, \
    register_storage_handler_cls


class SharedStorageIO(BytesStorageIO):
    storage_type = DataStorageDevice.SHARED_MEMORY
    filename = None

    def __init__(self, session_id, data_key, mode='w', shared_store=None,
                 nbytes=None, packed=False, compress=None, auto_register=True,
                 pin_token=None, handler=None):
        from .objectholder import SharedHolderActor

        super().__init__(session_id, data_key, mode=mode, handler=handler)
        self._shared_buf = None
        self._shared_store = shared_store
        self._offset = 0
        self._nbytes = nbytes
        self._holder_ref = self._storage_ctx.actor_ctx.actor_ref(SharedHolderActor.default_uid())
        self._compress = compress or dataserializer.CompressType.NONE
        self._packed = packed
        self._auto_register = auto_register
        self._pin_token = pin_token

        block_size = options.worker.copy_block_size

        if self.is_writable:
            self._shared_buf = shared_store.create(session_id, data_key, nbytes)
            if packed:
                self._buf = ArrowBufferIO(self._shared_buf, 'w', block_size=block_size)
            else:
                import pyarrow
                self._buf = pyarrow.FixedSizeBufferWriter(self._shared_buf)
                self._buf.set_memcopy_threads(6)
        elif self.is_readable:
            self._shared_buf = shared_store.get_buffer(session_id, data_key)
            if packed:
                self._buf = ArrowBufferIO(
                    self._shared_buf, 'r', compress_out=compress, block_size=block_size)
            else:
                self._mv = memoryview(self._shared_buf)
                self._nbytes = len(self._shared_buf)
        else:
            raise NotImplementedError

    def __del__(self):
        self._mv = None
        self._buf = self._shared_buf = None

    @property
    def nbytes(self):
        return self._nbytes

    def get_shared_buffer(self):
        return self._shared_buf

    def read(self, size=-1):
        if self._packed:
            return self._buf.read(size)
        else:
            if size < 0:
                size = self._nbytes
            right_pos = min(self._nbytes, self._offset + size)
            ret = self._mv[self._offset:right_pos]
            self._offset = right_pos
            return ret

    def write(self, d):
        return self._buf.write(d)

    def close(self, finished=True):
        if self._closed:
            return

        if self.is_writable and self._shared_buf is not None:
            self._shared_store.seal(self._session_id, self._data_key)
            if finished:
                if self._auto_register:
                    self._holder_ref.put_objects_by_keys(
                        self._session_id, [self._data_key], pin_token=self._pin_token)
            else:
                self._shared_store.delete(self._session_id, self._data_key)

        self._mv = None
        self._buf = self._shared_buf = None
        super().close(finished=finished)


class SharedStorageHandler(StorageHandler, BytesStorageMixin, ObjectStorageMixin,
                           SpillableStorageMixin):
    storage_type = DataStorageDevice.SHARED_MEMORY

    def __init__(self, storage_ctx, proc_id=None):
        from .objectholder import SharedHolderActor

        StorageHandler.__init__(self, storage_ctx, proc_id=proc_id)
        self._shared_store = storage_ctx.shared_store
        self._holder_ref = storage_ctx.promise_ref(SharedHolderActor.default_uid())

    @wrap_promised
    def create_bytes_reader(self, session_id, data_key, packed=False, packed_compression=None,
                            _promise=False):
        return SharedStorageIO(session_id, data_key, 'r', self._shared_store, packed=packed,
                               compress=packed_compression, handler=self)

    @wrap_promised
    def create_bytes_writer(self, session_id, data_key, total_bytes, packed=False,
                            packed_compression=None, auto_register=True, pin_token=None,
                            _promise=False):
        return SharedStorageIO(session_id, data_key, 'w', self._shared_store,
                               nbytes=total_bytes, packed=packed, auto_register=auto_register,
                               handler=self, pin_token=pin_token)

    @wrap_promised
    def get_objects(self, session_id, data_keys, serialize=False, _promise=False):
        if serialize:
            return [self._shared_store.get_buffer(session_id, k) for k in data_keys]
        else:
            return [self._shared_store.get(session_id, k) for k in data_keys]

    @wrap_promised
    def put_objects(self, session_id, data_keys, objs, sizes=None, shapes=None,
                    serialize=False, pin_token=None, _promise=False):
        succ_keys, succ_shapes = [], []
        affected_keys = []
        request_size, capacity = 0, 0

        objs = [self._deserial(obj) if serialize else obj for obj in objs]
        shapes = shapes or [getattr(obj, 'shape', None) for obj in objs]
        obj_refs = []
        obj = None
        try:
            for key, obj, shape in zip(data_keys, objs, shapes):
                try:
                    obj_refs.append(self._shared_store.put(session_id, key, obj))
                    succ_keys.append(key)
                    succ_shapes.append(shape)
                except StorageFull as ex:
                    affected_keys.extend(ex.affected_keys)
                    request_size += ex.request_size
                    capacity = ex.capacity
                except StorageDataExists:
                    if self.location in self.storage_ctx.get_data_locations(session_id, [key])[0]:
                        succ_keys.append(key)
                        succ_shapes.append(shape)
                    else:
                        raise
            self._holder_ref.put_objects_by_keys(session_id, succ_keys, shapes=succ_shapes, pin_token=pin_token)
            if affected_keys:
                raise StorageFull(request_size=request_size, capacity=capacity,
                                  affected_keys=affected_keys)
        finally:
            del obj
            objs[:] = []
            obj_refs[:] = []

    def load_from_bytes_io(self, session_id, data_keys, src_handler, pin_token=None):
        is_success = True
        shared_bufs = []
        failed_keys = set()
        storage_full_sizes = [0, 0]

        def _handle_writer_create_fail(key, reader, exc_info):
            nonlocal is_success
            reader.close()
            failed_keys.add(key)
            exc = exc_info[1]
            if isinstance(exc, StorageFull):
                storage_full_sizes[0] += exc.request_size
                storage_full_sizes[1] = exc.capacity
            else:
                is_success = False
                shared_bufs[:] = []
                self.pass_on_exc(reader.close, exc_info)

        def _on_file_close(key, _reader, writer, finished):
            if finished:
                if is_success:
                    shared_bufs.append(writer.get_shared_buffer())
            else:
                failed_keys.add(key)

        def _finalize_load(*exc_info):
            try:
                success_keys = [k for k in data_keys if k not in failed_keys]
                if success_keys:
                    self._holder_ref.put_objects_by_keys(session_id, success_keys, pin_token=pin_token)
                if exc_info:
                    raise exc_info[1].with_traceback(exc_info[2]) from None
                if failed_keys:
                    raise StorageFull(request_size=storage_full_sizes[0], capacity=storage_full_sizes[1],
                                      affected_keys=list(failed_keys))
            finally:
                shared_bufs[:] = []

        def _load_single_key(k):
            on_close = functools.partial(_on_file_close, k)
            handle_worker_create_fail = functools.partial(_handle_writer_create_fail, k)
            return src_handler.create_bytes_reader(session_id, k, _promise=True) \
                .then(lambda reader: self.create_bytes_writer(
                    session_id, k, reader.nbytes, auto_register=False, _promise=True)
                        .then(lambda writer: self._copy_bytes_data(reader, writer, on_close),
                              lambda *exc: handle_worker_create_fail(reader, exc)))

        def _fallback(*_):
            return promise.all_(_load_single_key(k) for k in data_keys) \
                .then(lambda *_: _finalize_load(), _finalize_load)

        return self.transfer_in_runner(session_id, data_keys, src_handler, _fallback)

    def load_from_object_io(self, session_id, data_keys, src_handler, pin_token=None):
        def _fallback(*_):
            ser_needed = src_handler.storage_type not in \
                         (DataStorageDevice.SHARED_MEMORY, DataStorageDevice.PROC_MEMORY)
            return self._batch_load_objects(
                session_id, data_keys,
                lambda k: src_handler.get_objects(session_id, k, serialize=ser_needed, _promise=True),
                serialize=ser_needed, pin_token=pin_token, batch_get=True)

        return self.transfer_in_runner(session_id, data_keys, src_handler, _fallback)

    def delete(self, session_id, data_keys, _tell=False):
        self._holder_ref.delete_objects(session_id, data_keys, _tell=_tell)
        self.unregister_data(session_id, data_keys, _tell=_tell)

    def spill_size(self, size, multiplier=1):
        return self._holder_ref.spill_size(size, multiplier, _promise=True)

    def lift_data_keys(self, session_id, data_keys):
        self._holder_ref.lift_data_keys(session_id, data_keys, _tell=True)

    def pin_data_keys(self, session_id, data_keys, token):
        return self._holder_ref.pin_data_keys(session_id, data_keys, token)

    def unpin_data_keys(self, session_id, data_keys, token, _tell=False):
        return self._holder_ref.unpin_data_keys(session_id, data_keys, token, _tell=_tell)


register_storage_handler_cls(DataStorageDevice.SHARED_MEMORY, SharedStorageHandler)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .concat import DataFrameConcat, concat
from .merge import join, merge, DataFrameShuffleMerge, DataFrameMergeAlign
from .append import DataFrameAppend, append


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE

    for cls in DATAFRAME_TYPE:
        setattr(cls, 'join', join)
        setattr(cls, 'merge', merge)

    for cls in DATAFRAME_TYPE + SERIES_TYPE:
        setattr(cls, 'append', append)


_install()
del _install



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...config import options
from ...core import OutputType
from ...serialize import BoolField
from ..operands import DataFrameOperand, DataFrameOperandMixin, DATAFRAME_TYPE


class DataFrameCheckNA(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.CHECK_NA

    _positive = BoolField('positive')
    _use_inf_as_na = BoolField('use_inf_as_na')

    def __init__(self, positive=None, use_inf_as_na=None, sparse=None, output_types=None, **kw):
        super().__init__(_positive=positive, _use_inf_as_na=use_inf_as_na, _sparse=sparse,
                         _output_types=output_types, **kw)

    @property
    def positive(self) -> bool:
        return self._positive

    @property
    def use_inf_as_na(self) -> bool:
        return self._use_inf_as_na

    def __call__(self, df):
        if isinstance(df, DATAFRAME_TYPE):
            self.output_types = [OutputType.dataframe]
        else:
            self.output_types = [OutputType.series]

        params = df.params.copy()
        if self.output_types[0] == OutputType.dataframe:
            params['dtypes'] = pd.Series([np.dtype('bool')] * len(df.dtypes),
                                         index=df.columns_value.to_pandas())
        else:
            params['dtype'] = np.dtype('bool')
        return self.new_tileable([df], **params)

    @classmethod
    def tile(cls, op: "DataFrameCheckNA"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_df.chunks:
            params = c.params.copy()
            if op.output_types[0] == OutputType.dataframe:
                params['dtypes'] = pd.Series([np.dtype('bool')] * len(c.dtypes),
                                             index=c.columns_value.to_pandas())
            else:
                params['dtype'] = np.dtype('bool')
            new_op = op.copy().reset_key()
            chunks.append(new_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        params = out_df.params.copy()
        params.update(dict(chunks=chunks, nsplits=in_df.nsplits))
        return new_op.new_tileables([in_df], **params)

    @classmethod
    def execute(cls, ctx, op: "DataFrameCheckNA"):
        in_data = ctx[op.inputs[0].key]
        try:
            pd.set_option('mode.use_inf_as_na', op.use_inf_as_na)
            if op.positive:
                ctx[op.outputs[0].key] = in_data.isna()
            else:
                ctx[op.outputs[0].key] = in_data.notna()
        finally:
            pd.reset_option('mode.use_inf_as_na')


def isna(df):
    """
    Detect missing values.

    Return a boolean same-sized object indicating if the values are NA.
    NA values, such as None or :attr:`numpy.NaN`, gets mapped to True
    values.

    Everything else gets mapped to False values. Characters such as empty
    strings ``''`` or :attr:`numpy.inf` are not considered NA values
    (unless you set ``pandas.options.mode.use_inf_as_na = True``).

    Returns
    -------
    DataFrame
        Mask of bool values for each element in DataFrame that
        indicates whether an element is not an NA value.

    See Also
    --------
    DataFrame.isnull : Alias of isna.
    DataFrame.notna : Boolean inverse of isna.
    DataFrame.dropna : Omit axes labels with missing values.
    isna : Top-level isna.

    Examples
    --------
    Show which entries in a DataFrame are NA.

    >>> import numpy as np
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'age': [5, 6, np.NaN],
    ...                    'born': [md.NaT, md.Timestamp('1939-05-27'),
    ...                             md.Timestamp('1940-04-25')],
    ...                    'name': ['Alfred', 'Batman', ''],
    ...                    'toy': [None, 'Batmobile', 'Joker']})
    >>> df.execute()
       age       born    name        toy
    0  5.0        NaT  Alfred       None
    1  6.0 1939-05-27  Batman  Batmobile
    2  NaN 1940-04-25              Joker

    >>> df.isna().execute()
         age   born   name    toy
    0  False   True  False   True
    1  False  False  False  False
    2   True  False  False  False

    Show which entries in a Series are NA.

    >>> ser = md.Series([5, 6, np.NaN])
    >>> ser.execute()
    0    5.0
    1    6.0
    2    NaN
    dtype: float64

    >>> ser.isna().execute()
    0    False
    1    False
    2     True
    dtype: bool
    """
    op = DataFrameCheckNA(positive=True, use_inf_as_na=options.dataframe.mode.use_inf_as_na)
    return op(df)


def notna(df):
    """
    Detect existing (non-missing) values.

    Return a boolean same-sized object indicating if the values are not NA.
    Non-missing values get mapped to True. Characters such as empty
    strings ``''`` or :attr:`numpy.inf` are not considered NA values
    (unless you set ``pandas.options.mode.use_inf_as_na = True``).
    NA values, such as None or :attr:`numpy.NaN`, get mapped to False
    values.

    Returns
    -------
    DataFrame
        Mask of bool values for each element in DataFrame that
        indicates whether an element is not an NA value.

    See Also
    --------
    DataFrame.notnull : Alias of notna.
    DataFrame.isna : Boolean inverse of notna.
    DataFrame.dropna : Omit axes labels with missing values.
    notna : Top-level notna.

    Examples
    --------
    Show which entries in a DataFrame are not NA.

    >>> import numpy as np
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'age': [5, 6, np.NaN],
    ...                    'born': [md.NaT, md.Timestamp('1939-05-27'),
    ...                             md.Timestamp('1940-04-25')],
    ...                    'name': ['Alfred', 'Batman', ''],
    ...                    'toy': [None, 'Batmobile', 'Joker']})
    >>> df.execute()
       age       born    name        toy
    0  5.0        NaT  Alfred       None
    1  6.0 1939-05-27  Batman  Batmobile
    2  NaN 1940-04-25              Joker

    >>> df.notna().execute()
         age   born  name    toy
    0   True  False  True  False
    1   True   True  True   True
    2  False   True  True   True

    Show which entries in a Series are not NA.

    >>> ser = md.Series([5, 6, np.NaN])
    >>> ser.execute()
    0    5.0
    1    6.0
    2    NaN
    dtype: float64

    >>> ser.notna().execute()
    0     True
    1     True
    2    False
    dtype: bool
    """
    op = DataFrameCheckNA(positive=False, use_inf_as_na=options.dataframe.mode.use_inf_as_na)
    return op(df)


isnull = isna
notnull = notna



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import NDArrayField, StringField, Float64Field
from ...config import options
from ..utils import decide_chunk_sizes, gen_random_seeds
from ..array_utils import array_module, device
from .core import TensorRandomOperandMixin, TensorDistribution, TENSOR_CHUNK_TYPE


class TensorMultivariateNormal(TensorDistribution, TensorRandomOperandMixin):
    __slots__ = '_mean', '_cov', '_size', '_check_valid', '_tol'
    _op_type_ = OperandDef.RAND_MULTIVARIATE_NORMAL

    _mean = NDArrayField('mean')
    _cov = NDArrayField('cov')
    _check_valid = StringField('check_valid')
    _tol = Float64Field('tol')
    _func_name = 'multivariate_normal'

    def __init__(self, mean=None, cov=None, size=None, check_valid=None, tol=None,
                 state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super().__init__(_mean=mean, _cov=cov, _size=size, _check_valid=check_valid,
                         _tol=tol, _state=state, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def mean(self):
        return self._mean

    @property
    def cov(self):
        return self._cov

    @property
    def check_valid(self):
        return self._check_valid

    @property
    def tol(self):
        return self._tol

    def __call__(self, chunk_size=None):
        N = self._mean.size
        if self._size is None:
            shape = (N,)
        else:
            try:
                shape = tuple(self._size) + (N,)
            except TypeError:
                shape = (self._size, N)

        return self.new_tensor(None, shape, raw_chunk_size=chunk_size)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        chunk_size = tensor.extra_params.raw_chunk_size or options.chunk_size
        nsplits = decide_chunk_sizes(tensor.shape[:-1], chunk_size, tensor.dtype.itemsize) + ((tensor.shape[-1],),)

        mean_chunk = op.mean.chunks[0] if hasattr(op.mean, 'chunks') else op.mean
        cov_chunk = op.cov.chunks[0] if hasattr(op.cov, 'chunks') else op.cov

        idxes = list(itertools.product(*[range(len(s)) for s in nsplits]))
        seeds = gen_random_seeds(len(idxes), op.state)

        out_chunks = []
        for seed, out_idx, shape in zip(seeds, idxes, itertools.product(*nsplits)):
            chunk_op = op.copy().reset_key()
            chunk_op._state = None
            chunk_op._seed = seed
            chunk_op._size = shape[:-1]
            out_chunk = chunk_op.new_chunk([mean_chunk, cov_chunk], shape=shape, index=out_idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape,
                                  chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def execute(cls, ctx, op):
        xp = array_module(op.gpu)
        if xp is np:
            device_id = -1
        else:
            device_id = op.device or 0

        with device(device_id):
            rs = xp.random.RandomState(op.seed)

            args = []
            for k in op.args:
                val = getattr(op, k, None)
                if isinstance(val, TENSOR_CHUNK_TYPE):
                    args.append(ctx[val.key])
                else:
                    args.append(val)
            mean, cov = args[:2]
            kw = {}
            if args[2] is not None:
                kw['size'] = args[2]
            if args[3] is not None:
                kw['check_valid'] = args[3]
            if args[4] is not None:
                kw['tol'] = args[4]

            try:
                res = rs.multivariate_normal(mean, cov, **kw)
                if xp is not np:
                    ctx[op.outputs[0].key] = xp.asarray(res)
                else:
                    ctx[op.outputs[0].key] = res
            except AttributeError:
                if xp is not np:
                    # cupy cannot generate data, fallback to numpy first
                    rs = np.random.RandomState(op.seed)
                    res = rs.multivariate_normal(mean, cov, **kw)
                    ctx[op.outputs[0].key] = xp.asarray(res)
                else:
                    raise


def multivariate_normal(random_state, mean, cov, size=None, check_valid=None, tol=None,
                        chunk_size=None, gpu=None, dtype=None):
    """
    Draw random samples from a multivariate normal distribution.

    The multivariate normal, multinormal or Gaussian distribution is a
    generalization of the one-dimensional normal distribution to higher
    dimensions.  Such a distribution is specified by its mean and
    covariance matrix.  These parameters are analogous to the mean
    (average or "center") and variance (standard deviation, or "width,"
    squared) of the one-dimensional normal distribution.

    Parameters
    ----------
    mean : 1-D array_like, of length N
        Mean of the N-dimensional distribution.
    cov : 2-D array_like, of shape (N, N)
        Covariance matrix of the distribution. It must be symmetric and
        positive-semidefinite for proper sampling.
    size : int or tuple of ints, optional
        Given a shape of, for example, ``(m,n,k)``, ``m*n*k`` samples are
        generated, and packed in an `m`-by-`n`-by-`k` arrangement.  Because
        each sample is `N`-dimensional, the output shape is ``(m,n,k,N)``.
        If no shape is specified, a single (`N`-D) sample is returned.
    check_valid : { 'warn', 'raise', 'ignore' }, optional
        Behavior when the covariance matrix is not positive semidefinite.
    tol : float, optional
        Tolerance when checking the singular values in covariance matrix.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor
        The drawn samples, of shape *size*, if that was provided.  If not,
        the shape is ``(N,)``.

        In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
        value drawn from the distribution.

    Notes
    -----
    The mean is a coordinate in N-dimensional space, which represents the
    location where samples are most likely to be generated.  This is
    analogous to the peak of the bell curve for the one-dimensional or
    univariate normal distribution.

    Covariance indicates the level to which two variables vary together.
    From the multivariate normal distribution, we draw N-dimensional
    samples, :math:`X = [x_1, x_2, ... x_N]`.  The covariance matrix
    element :math:`C_{ij}` is the covariance of :math:`x_i` and :math:`x_j`.
    The element :math:`C_{ii}` is the variance of :math:`x_i` (i.e. its
    "spread").

    Instead of specifying the full covariance matrix, popular
    approximations include:

      - Spherical covariance (`cov` is a multiple of the identity matrix)
      - Diagonal covariance (`cov` has non-negative elements, and only on
        the diagonal)

    This geometrical property can be seen in two dimensions by plotting
    generated data-points:

    >>> mean = [0, 0]
    >>> cov = [[1, 0], [0, 100]]  # diagonal covariance

    Diagonal covariance means that points are oriented along x or y-axis:

    >>> import matplotlib.pyplot as plt
    >>> import mars.tensor as mt
    >>> x, y = mt.random.multivariate_normal(mean, cov, 5000).T
    >>> plt.plot(x.execute(), y.execute(), 'x')
    >>> plt.axis('equal')
    >>> plt.show()

    Note that the covariance matrix must be positive semidefinite (a.k.a.
    nonnegative-definite). Otherwise, the behavior of this method is
    undefined and backwards compatibility is not guaranteed.

    References
    ----------
    .. [1] Papoulis, A., "Probability, Random Variables, and Stochastic
           Processes," 3rd ed., New York: McGraw-Hill, 1991.
    .. [2] Duda, R. O., Hart, P. E., and Stork, D. G., "Pattern
           Classification," 2nd ed., New York: Wiley, 2001.

    Examples
    --------
    >>> mean = (1, 2)
    >>> cov = [[1, 0], [0, 1]]
    >>> x = mt.random.multivariate_normal(mean, cov, (3, 3))
    >>> x.shape
    (3, 3, 2)

    The following is probably true, given that 0.6 is roughly twice the
    standard deviation:

    >>> list(((x[0,0,:] - mean) < 0.6).execute())
    [True, True]
    """
    mean = np.asarray(mean)
    cov = np.asarray(cov)

    if mean.ndim != 1:
        raise ValueError('mean must be 1 dimensional')
    if cov.ndim != 2:
        raise ValueError('cov must be 1 dimensional')
    if len(set(mean.shape + cov.shape)) != 1:
        raise ValueError('mean and cov must have same length')

    if dtype is None:
        small_kw = {}
        if check_valid:
            small_kw['check_valid'] = check_valid
        if tol:
            small_kw['tol'] = tol
        dtype = np.random.multivariate_normal(mean, cov, size=(0,), **small_kw).dtype

    size = random_state._handle_size(size)
    op = TensorMultivariateNormal(mean=mean, cov=cov, size=size, check_valid=check_valid,
                                  tol=tol, state=random_state.to_numpy(), gpu=gpu, dtype=dtype)
    return op(chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

from .. import kvstore
from ..config import options
from .session import SessionManagerActor
from .resource import ResourceActor
from .chunkmeta import ChunkMetaActor
from .kvstore import KVStoreActor
from .node_info import NodeInfoActor
from .utils import SchedulerClusterInfoActor


logger = logging.getLogger(__name__)


class SchedulerService(object):
    def __init__(self, **kwargs):
        self._cluster_info_ref = None
        self._session_manager_ref = None
        self._assigner_ref = None
        self._resource_ref = None
        self._chunk_meta_ref = None
        self._kv_store_ref = None
        self._node_info_ref = None
        self._result_receiver_ref = None

        options.scheduler.enable_failover = not (kwargs.pop('disable_failover', None) or False)

        if kwargs:  # pragma: no cover
            raise TypeError(f'Keyword arguments {kwargs!r} cannot be recognized.')

    def start(self, endpoint, discoverer, pool, distributed=True):
        """
        there are two way to start a scheduler
        1) if options.kv_store is specified as an etcd address, the endpoint will be written
        into kv-storage to indicate that this scheduler is one the schedulers,
        and the etcd is used as a service discover.
        2) if options.kv_store is not an etcd address, there will be only one scheduler
        """
        kv_store = kvstore.get(options.kv_store)
        kv_store.write(f'/schedulers/{endpoint}', dir=True)

        if not isinstance(kv_store, kvstore.LocalKVStore):
            # set etcd as service discover
            logger.info('Mars Scheduler started with kv store %s.', options.kv_store)
            # create KVStoreActor when there is a distributed KV store
            self._kv_store_ref = pool.create_actor(KVStoreActor, uid=KVStoreActor.default_uid())
        else:
            # single scheduler
            logger.info('Mars Scheduler started in standalone mode.')

        # create ClusterInfoActor
        self._cluster_info_ref = pool.create_actor(
            SchedulerClusterInfoActor, discoverer, distributed=distributed,
            uid=SchedulerClusterInfoActor.default_uid())
        # create ChunkMetaActor
        self._chunk_meta_ref = pool.create_actor(ChunkMetaActor, uid=ChunkMetaActor.default_uid())
        # create SessionManagerActor
        self._session_manager_ref = pool.create_actor(
            SessionManagerActor, uid=SessionManagerActor.default_uid())
        # create ResourceActor
        self._resource_ref = pool.create_actor(ResourceActor, uid=ResourceActor.default_uid())
        # create NodeInfoActor
        self._node_info_ref = pool.create_actor(NodeInfoActor, uid=NodeInfoActor.default_uid())
        kv_store.write(f'/schedulers/{endpoint}/meta',
                       json.dumps(self._resource_ref.get_workers_meta()))

    def stop(self, pool):
        pool.destroy_actor(self._resource_ref)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import KeyField
from .... import tensor as mt
from ....utils import recursive_tile
from ....tensor.core import TensorOrder
from ...preprocessing import normalize
from .core import PairwiseDistances


class CosineDistances(PairwiseDistances):
    _op_type_ = OperandDef.PAIRWISE_COSINE_DISTANCES

    _x = KeyField('x')
    _y = KeyField('y')

    def __init__(self, x=None, y=None, dtype=None, gpu=None, **kw):
        super().__init__(_x=x, _y=y, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._x = self._inputs[0]
        self._y = self._inputs[1]

    def __call__(self, x, y=None):
        x, y = self.check_pairwise_arrays(x, y)
        return self.new_tensor([x, y], shape=(x.shape[0], y.shape[0]),
                               order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        x, y = op.x, op.y
        if x is y:
            S = cosine_similarity(x)
        else:
            S = cosine_similarity(x, y)
        S = (S * -1) + 1
        S = mt.clip(S, 0, 2)
        if x is y:
            mt.fill_diagonal(S, 0.0)
        return [recursive_tile(S)]


def cosine_similarity(X, Y=None, dense_output=True):
    """Compute cosine similarity between samples in X and Y.

    Cosine similarity, or the cosine kernel, computes similarity as the
    normalized dot product of X and Y:

        K(X, Y) = <X, Y> / (||X||*||Y||)

    On L2-normalized data, this function is equivalent to linear_kernel.

    Read more in the :ref:`User Guide <cosine_similarity>`.

    Parameters
    ----------
    X : Tensor or sparse tensor, shape: (n_samples_X, n_features)
        Input data.

    Y : Tensor or sparse tensor, shape: (n_samples_Y, n_features)
        Input data. If ``None``, the output will be the pairwise
        similarities between all samples in ``X``.

    dense_output : boolean (optional), default True
        Whether to return dense output even when the input is sparse. If
        ``False``, the output is sparse if both input tensors are sparse.

    Returns
    -------
    kernel matrix : Tensor
        A tensor with shape (n_samples_X, n_samples_Y).
    """
    X, Y = PairwiseDistances.check_pairwise_arrays(X, Y)

    X_normalized = normalize(X, copy=True)
    if X is Y:
        Y_normalized = X_normalized
    else:
        Y_normalized = normalize(Y, copy=True)

    K = X_normalized.dot(Y_normalized.T)
    if dense_output:
        K = K.todense()
    return K


def cosine_distances(X, Y=None):
    """Compute cosine distance between samples in X and Y.

    Cosine distance is defined as 1.0 minus the cosine similarity.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    X : array_like, sparse matrix
        with shape (n_samples_X, n_features).

    Y : array_like, sparse matrix (optional)
        with shape (n_samples_Y, n_features).

    Returns
    -------
    distance matrix : Tensor
        A tensor with shape (n_samples_X, n_samples_Y).

    See also
    --------
    mars.learn.metrics.pairwise.cosine_similarity
    mars.tensor.spatial.distance.cosine : dense matrices only
    """
    op = CosineDistances(x=X, y=Y, dtype=np.dtype(np.float64))
    return op(X, y=Y)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ... import tensor as mt
from ...serialize import ValueType, KeyField, ListField
from ...utils import recursive_tile
from ..core import SERIES_TYPE
from ..initializer import DataFrame, Series
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, build_empty_df


class DataFrameDescribe(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DESCRIBE

    _input = KeyField('input')
    _percentiles = ListField('percentiles', ValueType.float64)
    _include = ListField('include')
    _exclude = ListField('exclude')

    def __init__(self, percentiles=None, include=None, exclude=None,
                 output_types=None, **kw):
        super().__init__(_percentiles=percentiles, _include=include,
                         _exclude=exclude, _output_types=output_types, **kw)

    @property
    def input(self):
        return self._input

    @property
    def percentiles(self):
        return self._percentiles

    @property
    def include(self):
        return self._include

    @property
    def exclude(self):
        return self._exclude

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, df_or_series):
        if isinstance(df_or_series, SERIES_TYPE):
            if not np.issubdtype(df_or_series.dtype, np.number):
                raise NotImplementedError('non-numeric type is not supported for now')
            test_series = pd.Series([], dtype=df_or_series.dtype).describe(
                percentiles=self._percentiles, include=self._include, exclude=self._exclude)
            return self.new_series([df_or_series], shape=(len(test_series),),
                                   dtype=test_series.dtype,
                                   index_value=parse_index(test_series.index, store_data=True))
        else:
            test_inp_df = build_empty_df(df_or_series.dtypes)
            test_df = test_inp_df.describe(
                percentiles=self._percentiles, include=self._include, exclude=self._exclude)
            for dtype in test_df.dtypes:
                if not np.issubdtype(dtype, np.number):
                    raise NotImplementedError('non-numeric type is not supported for now')
            return self.new_dataframe([df_or_series], shape=test_df.shape, dtypes=test_df.dtypes,
                                      index_value=parse_index(test_df.index, store_data=True),
                                      columns_value=parse_index(test_df.columns, store_data=True))

    @classmethod
    def tile(cls, op):
        inp = op.input

        if len(inp.chunks) == 1:
            return cls._tile_one_chunk(op)

        if isinstance(inp, SERIES_TYPE):
            return cls._tile_series(op)
        else:
            return cls._tile_dataframe(op)

    @classmethod
    def _tile_one_chunk(cls, op):
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        chunk_params = out.params.copy()
        chunk_params['index'] = (0,) * out.ndim
        out_chunk = chunk_op.new_chunk([op.input.chunks[0]], kws=[chunk_params])

        new_op = op.copy()
        params = out.params.copy()
        params['chunks'] = [out_chunk]
        params['nsplits'] = tuple((s,) for s in out.shape)
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _tile_series(cls, op):
        series = Series(op.input)
        out = op.outputs[0]
        index = out.index_value.to_pandas()
        # ['count', 'mean', 'std', 'min', {percentiles}, 'max']
        names = index.tolist()

        values = [None] * 6
        for i, agg in enumerate(names[:4]):
            values[i] = mt.atleast_1d(getattr(series, agg)())
        values[-1] = mt.atleast_1d(getattr(series, names[-1])())
        values[4] = series.quantile(op.percentiles).to_tensor()

        t = mt.concatenate(values).rechunk(len(names))
        ret = Series(t, index=index, name=series.name)
        return [recursive_tile(ret)]

    @classmethod
    def _tile_dataframe(cls, op):
        df = DataFrame(op.input)
        out = op.outputs[0]
        index = out.index_value.to_pandas()
        dtypes = out.dtypes
        columns = dtypes.index.tolist()
        # ['count', 'mean', 'std', 'min', {percentiles}, 'max']
        names = index.tolist()

        df = df[columns]

        values = [None] * 6
        for i, agg in enumerate(names[:4]):
            values[i] = getattr(df, agg)().to_tensor()[None, :]
        values[-1] = getattr(df, names[-1])().to_tensor()[None, :]
        values[4] = df.quantile(op.percentiles).to_tensor()

        t = mt.concatenate(values).rechunk((len(index), len(columns)))
        ret = DataFrame(t, index=index, columns=columns)
        return [recursive_tile(ret)]

    @classmethod
    def execute(cls, ctx, op):
        df_or_series = ctx[op.input.key]

        ctx[op.outputs[0].key] = df_or_series.describe(
            percentiles=op.percentiles, include=op.include, exclude=op.exclude)


def describe(df_or_series, percentiles=None, include=None, exclude=None):
    if percentiles is not None:
        for p in percentiles:
            if p < 0 or p > 1:
                raise ValueError('percentiles should all be in the interval [0, 1]. '
                                 'Try [{0:.3f}] instead.'.format(p / 100))
    if percentiles is None:
        percentiles = [0.25, 0.5, 0.75]
    if not percentiles:
        percentiles = [0.5]

    op = DataFrameDescribe(percentiles=percentiles, include=include, exclude=exclude)
    return op(df_or_series)
